from dagster import (
    DagsterInvariantViolationError,
    ExecutionContext,
    PipelineDefinition,
    PipelineContextDefinition,
    Result,
    SolidDefinition,
    check,
    config,
    execute_pipeline,
)


def execute_single_solid(context, solid, environment=None, throw_on_error=True):
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(solid, 'solid', SolidDefinition)
    environment = check.opt_inst_param(
        environment,
        'environment',
        config.Environment,
        config.Environment(),
    )
    check.bool_param(throw_on_error, 'throw_on_error')

    single_solid_environment = config.Environment(
        expectations=environment.expectations,
        context=environment.context,
        solids={solid.name: environment.solids[solid.name]}
        if solid.name in environment.solids else None
    )

    pipeline_result = execute_pipeline(
        PipelineDefinition(
            solids=[solid],
            context_definitions=PipelineContextDefinition.passthrough_context_definition(context),
        ),
        environment=single_solid_environment,
    )

    return pipeline_result


def single_output_transform(name, inputs, transform_fn, output, description=None):
    '''It is commmon to want a Solid that has only inputs, a single output (with the default
    name), and no config. So this is a helper function to do that. This transform function
    must return the naked return value (as opposed to a Result object).

    Args:
        name (str): Name of the solid.
        inputs (List[InputDefinition]): Inputs of solid.
        transform_fn (callable):
            Callable with the signature
            (context: ExecutionContext, inputs: Dict[str, Any]) : Any
        output (OutputDefinition): Output of the solid.
        description (str): Descripion of the solid.

    Returns:
        SolidDefinition:

    Example:

        .. code-block:: python

            single_output_transform(
                'add_one',
                inputs=InputDefinition('num', types.Int),
                output=OutputDefinition(types.Int),
                transform_fn=lambda context, inputs: inputs['num'] + 1
            )

    '''

    def _new_transform_fn(info, inputs):
        value = transform_fn(info.context, inputs)
        if isinstance(value, Result):
            raise DagsterInvariantViolationError(
                '''Single output transform Solid {name} returned a Result. Just return
                value directly without wrapping it in Result'''
            )
        yield Result(value=value)

    return SolidDefinition(
        name=name,
        inputs=inputs,
        transform_fn=_new_transform_fn,
        outputs=[output],
        description=description,
    )


from dagster import (RepositoryDefinition)
from dagster.dagster_examples.pandas_hello_world.pipeline import define_success_pipeline


def define_example_repository():
    return RepositoryDefinition(
        name='example_repo',
        pipeline_dict={'pandas_hello_world': define_success_pipeline},
    )



import pandas as pd

from dagster import (
    ConfigDefinition,
    DependencyDefinition,
    Field,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    Result,
    SolidDefinition,
    config,
    execute_pipeline,
    types,
)

from dagster.core.test_utils import single_output_transform

from dagster.utils import script_relative_path


def define_read_csv_solid(name):
    def _t_fn(info, _inputs):
        yield Result(pd.read_csv(info.config['path']))

    return SolidDefinition(
        name=name,
        inputs=[],
        outputs=[OutputDefinition()],
        config_def=ConfigDefinition.config_dict({
            'path': Field(types.Path)
        }),
        transform_fn=_t_fn
    )


def define_to_csv_solid(name):
    def _t_fn(info, inputs):
        inputs['df'].to_csv(info.config['path'], index=False)

    return SolidDefinition(
        name=name,
        inputs=[InputDefinition('df')],
        outputs=[],
        config_def=ConfigDefinition.config_dict({
            'path': Field(types.Path)
        }),
        transform_fn=_t_fn,
    )


def test_hello_world_pipeline_no_api():
    def hello_world_transform_fn(_context, inputs):
        num_df = inputs['num_df']
        num_df['sum'] = num_df['num1'] + num_df['num2']
        return num_df

    read_csv_solid = define_read_csv_solid('read_csv_solid')

    hello_world = single_output_transform(
        name='hello_world',
        inputs=[InputDefinition('num_df')],
        transform_fn=hello_world_transform_fn,
        output=OutputDefinition(),
    )

    pipeline = PipelineDefinition(
        solids=[read_csv_solid, hello_world],
        dependencies={
            'hello_world': {
                'num_df': DependencyDefinition('read_csv_solid'),
            },
        }
    )

    pipeline_result = execute_pipeline(
        pipeline,
        config.Environment(
            solids={
                'read_csv_solid': config.Solid({
                    'path': script_relative_path('num.csv'),
                }),
            },
        ),
    )

    assert pipeline_result.success

    result = pipeline_result.result_for_solid('hello_world')

    assert result.transformed_value().to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }


def create_hello_world_solid_composed_pipeline():
    def transform_fn(_context, inputs):
        num_df = inputs['num_df']
        num_df['sum'] = num_df['num1'] + num_df['num2']
        return num_df

    hello_world = single_output_transform(
        name='hello_world',
        inputs=[InputDefinition('num_df')],
        transform_fn=transform_fn,
        output=OutputDefinition(),
    )

    return PipelineDefinition(
        solids=[define_read_csv_solid('read_hello_world'), hello_world],
        dependencies={'hello_world': {
            'num_df': DependencyDefinition('read_hello_world')
        }}
    )


def test_hello_world_composed():
    pipeline = create_hello_world_solid_composed_pipeline()

    pipeline_result = execute_pipeline(
        pipeline,
        environment=config.Environment(
            solids={
                'read_hello_world': config.Solid({
                    'path': script_relative_path('num.csv')
                }),
            },
        ),
    )

    assert pipeline_result.success

    result = pipeline_result.result_for_solid('hello_world')

    assert result.success

    assert result.transformed_value().to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }


def test_pandas_hello_no_library():
    def solid_one_transform(_context, inputs):
        num_df = inputs['num_df']
        num_df['sum'] = num_df['num1'] + num_df['num2']
        return num_df

    solid_one = single_output_transform(
        name='solid_one',
        inputs=[InputDefinition(name='num_df')],
        transform_fn=solid_one_transform,
        output=OutputDefinition(),
    )

    def solid_two_transform(_context, inputs):
        sum_df = inputs['sum_df']
        sum_df['sum_sq'] = sum_df['sum'] * sum_df['sum']
        return sum_df

    solid_two = single_output_transform(
        name='solid_two',
        inputs=[InputDefinition(name='sum_df')],
        transform_fn=solid_two_transform,
        output=OutputDefinition(),
    )

    pipeline = PipelineDefinition(
        solids=[define_read_csv_solid('read_one'), solid_one, solid_two],
        dependencies={
            'solid_one': {
                'num_df': DependencyDefinition('read_one'),
            },
            'solid_two': {
                'sum_df': DependencyDefinition('solid_one'),
            },
        }
    )

    environment = config.Environment(
        solids={
            'read_one': config.Solid({
                'path': script_relative_path('num.csv')
            }),
        }
    )

    execute_pipeline_result = execute_pipeline(
        pipeline,
        environment=environment,
    )

    assert execute_pipeline_result.result_for_solid('solid_two'
                                                    ).transformed_value().to_dict('list') == {
                                                        'num1': [1, 3],
                                                        'num2': [2, 4],
                                                        'sum': [3, 7],
                                                        'sum_sq': [9, 49],
                                                    }

    sum_sq_out_path = '/tmp/sum_sq.csv'
    import os
    if os.path.exists(sum_sq_out_path):
        os.remove(sum_sq_out_path)

    sum_sq_path_args = {'path': '/tmp/sum_sq.csv'}
    environment_two = config.Environment(
        solids={
            'read_one': config.Solid({
                'path': script_relative_path('num.csv')
            }),
            'write_two': config.Solid(sum_sq_path_args),
        },
    )

    pipeline_two = PipelineDefinition(
        solids=[
            define_read_csv_solid('read_one'),
            solid_one,
            solid_two,
            define_to_csv_solid('write_two'),
        ],
        dependencies={
            'solid_one': {
                'num_df': DependencyDefinition('read_one'),
            },
            'solid_two': {
                'sum_df': DependencyDefinition('solid_one'),
            },
            'write_two': {
                'df': DependencyDefinition('solid_two'),
            }
        }
    )

    execute_pipeline(pipeline_two, environment=environment_two)

    sum_sq_df = pd.read_csv('/tmp/sum_sq.csv')

    assert sum_sq_df.to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_sq': [9, 49],
    }



'''
Naming conventions:

For public functions:

execute_*

These represent functions which do purely in-memory compute. They will evaluate expectations
the core transform, and exercise all logging and metrics tracking (outside of outputs), but they
will not invoke *any* outputs (and their APIs don't allow the user to).


'''

# too many lines
# pylint: disable=C0302

from contextlib import contextmanager
import itertools

import six

from dagster import (
    check,
    config,
)

from .definitions import (
    DEFAULT_OUTPUT,
    ContextCreationExecutionInfo,
    ExecutionGraph,
    PipelineDefinition,
    SolidDefinition,
)

from .errors import (
    DagsterEvaluateValueError,
    DagsterInvariantViolationError,
    DagsterTypeError,
    DagsterUserCodeExecutionError,
)

from .compute_nodes import (
    ComputeNodeExecutionInfo,
    ComputeNodeResult,
    ComputeNodeTag,
    create_compute_node_graph,
    execute_compute_nodes,
)

from .execution_context import ExecutionContext


class PipelineExecutionResult(object):
    '''Result of execution of the whole pipeline. Returned eg by :py:function:`execute_pipeline`.

    Attributes:
        pipeline (PipelineDefinition): Pipeline that was executed
        context (ExecutionContext): ExecutionContext of that particular Pipeline run.
        result_list (list[SolidExecutionResult]): List of results for each pipeline solid.
    '''

    def __init__(
        self,
        pipeline,
        context,
        result_list,
    ):
        self.pipeline = check.inst_param(pipeline, 'pipeline', PipelineDefinition)
        self.context = check.inst_param(context, 'context', ExecutionContext)
        self.result_list = check.list_param(
            result_list, 'result_list', of_type=SolidExecutionResult
        )

    @property
    def success(self):
        '''Whether the pipeline execution was successful at all steps'''
        return all([result.success for result in self.result_list])

    def result_for_solid(self, name):
        '''Get a :py:class:`SolidExecutionResult` for a given solid name.

        Returns:
          SolidExecutionResult
        '''
        check.str_param(name, 'name')

        if not self.pipeline.has_solid(name):
            raise DagsterInvariantViolationError(
                'Try to get result for solid {name} in {pipeline}. No such solid.'.format(
                    name=name,
                    pipeline=self.pipeline.display_name,
                )
            )

        for result in self.result_list:
            if result.solid.name == name:
                return result

        raise DagsterInvariantViolationError(
            'Did not find result for solid {name} in pipeline execution result'.format(name=name)
        )


class SolidExecutionResult(object):
    '''Execution result for one solid of the pipeline.

    Attributes:
      context (ExecutionContext): ExecutionContext of that particular Pipeline run.
      solid (SolidDefinition): Solid for which this result is
    '''

    def __init__(self, context, solid, input_expectations, transforms, output_expectations):
        self.context = check.inst_param(context, 'context', ExecutionContext)
        self.solid = check.inst_param(solid, 'solid', SolidDefinition)
        self.input_expectations = check.list_param(
            input_expectations, 'input_expectations', ComputeNodeResult
        )
        self.output_expectations = check.list_param(
            output_expectations, 'output_expectations', ComputeNodeResult
        )
        self.transforms = check.list_param(transforms, 'transforms', ComputeNodeResult)

    @staticmethod
    def from_results(context, results):
        results = check.list_param(results, 'results', ComputeNodeResult)
        if results:
            input_expectations = []
            output_expectations = []
            transforms = []

            for result in results:
                if result.tag == ComputeNodeTag.INPUT_EXPECTATION:
                    input_expectations.append(result)
                elif result.tag == ComputeNodeTag.OUTPUT_EXPECTATION:
                    output_expectations.append(result)
                elif result.tag == ComputeNodeTag.TRANSFORM:
                    transforms.append(result)

            return SolidExecutionResult(
                context=context,
                solid=results[0].compute_node.solid,
                input_expectations=input_expectations,
                output_expectations=output_expectations,
                transforms=transforms,
            )
        else:
            check.failed("Cannot create SolidExecutionResult from empty list")

    @property
    def success(self):
        '''Whether the solid execution was successful'''
        return all(
            [
                result.success for result in
                itertools.chain(self.input_expectations, self.output_expectations, self.transforms)
            ]
        )

    @property
    def transformed_values(self):
        '''Return dictionary of transformed results, with keys being output names.
        Returns None if execution result isn't a success.'''
        if self.success and self.transforms:
            return {
                result.success_data.output_name: result.success_data.value
                for result in self.transforms
            }
        else:
            return None

    def transformed_value(self, output_name=DEFAULT_OUTPUT):
        '''Returns transformed value either for DEFAULT_OUTPUT or for the output
        given as output_name. Returns None if execution result isn't a success'''
        check.str_param(output_name, 'output_name')

        if not self.solid.has_output(output_name):
            raise DagsterInvariantViolationError(
                '{output_name} not defined in solid {solid}'.format(
                    output_name=output_name,
                    solid=self.solid.name,
                )
            )

        if self.success:
            for result in self.transforms:
                if result.success_data.output_name == output_name:
                    return result.success_data.value
            raise DagsterInvariantViolationError(
                'Did not find result {output_name} in solid {self.solid.name} execution result'.
                format(output_name=output_name, self=self)
            )
        else:
            return None

    def reraise_user_error(self):
        if not self.success:
            for result in itertools.chain(
                self.input_expectations, self.output_expectations, self.transforms
            ):
                if not result.success:
                    if isinstance(
                        result.failure_data.dagster_error,
                        DagsterUserCodeExecutionError,
                    ):
                        six.reraise(*result.failure_data.dagster_error.original_exc_info)
                    else:
                        raise result.failure_data.dagster_error

    @property
    def dagster_error(self):
        '''Returns exception that happened during this solid's execution, if any'''
        for result in itertools.chain(
            self.input_expectations, self.output_expectations, self.transforms
        ):
            if not result.success:
                return result.failure_data.dagster_error


def _wrap_in_yield(context_or_generator):
    if isinstance(context_or_generator, ExecutionContext):

        def _wrap():
            yield context_or_generator

        return _wrap()

    return context_or_generator


def _validate_environment(environment, pipeline):
    context_name = environment.context.name

    if context_name not in pipeline.context_definitions:
        avaiable_context_keys = list(pipeline.context_definitions.keys())
        raise DagsterInvariantViolationError(
            'Context {context_name} not found in '.format(context_name=context_name) + \
            'pipeline definiton. Available contexts {avaiable_context_keys}'.format(
                avaiable_context_keys=repr(avaiable_context_keys),
            )
        )

    for solid_name in environment.solids.keys():
        if not pipeline.has_solid(solid_name):
            available_solids = [s.name for s in pipeline.solids]
            raise DagsterInvariantViolationError(
                'Solid {solid_name} specified in config for pipeline {pipeline} not found.'.format(
                    solid_name=solid_name,
                    pipeline=pipeline.display_name,
                ) + \
                ' Available solids in pipeline are {solids}.'.format(solids=available_solids)
            )


def _create_config_value(config_type, config_input):
    try:
        return config_type.evaluate_value(config_input)
    except DagsterEvaluateValueError as e:
        raise DagsterTypeError(
            'Invalid config value: {error_msg}'.format(error_msg=','.join(e.args))
        )


@contextmanager
def yield_context(pipeline, environment):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(environment, 'environment', config.Environment)

    _validate_environment(environment, pipeline)

    context_name = environment.context.name
    context_definition = pipeline.context_definitions[context_name]
    config_type = context_definition.config_def.config_type

    config_value = _create_config_value(config_type, environment.context.config)
    context_or_generator = context_definition.context_fn(
        ContextCreationExecutionInfo(
            config=config_value,
            pipeline_def=pipeline,
        )
    )
    return _wrap_in_yield(context_or_generator)


def execute_pipeline_iterator(pipeline, environment):
    '''Returns iterator that yields :py:class:`SolidExecutionResult` for each
    solid executed in the pipeline.

    Parameters:
      pipeline (PipelineDefinition): pipeline to run
      execution (ExecutionContext): execution context of the run
    '''
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.inst_param(environment, 'enviroment', config.Environment)

    execution_graph = ExecutionGraph.from_pipeline(pipeline)
    with yield_context(pipeline, environment) as context:
        return _execute_graph_iterator(context, execution_graph, environment)


def _execute_graph_iterator(context, execution_graph, environment):
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(execution_graph, 'execution_graph', ExecutionGraph)
    check.inst_param(environment, 'environent', config.Environment)

    cn_graph = create_compute_node_graph(
        ComputeNodeExecutionInfo(
            context,
            execution_graph,
            environment,
        ),
    )

    cn_nodes = list(cn_graph.topological_nodes())

    if not cn_nodes:
        context.debug(
            'Pipeline {pipeline} has no nodes and no execution will happen'.format(
                pipeline=execution_graph.pipeline.display_name
            )
        )
        return

    context.debug(
        'About to execute the compute node graph in the following order {order}'.format(
            order=[cn.friendly_name for cn in cn_nodes]
        )
    )

    check.invariant(len(cn_nodes[0].node_inputs) == 0)

    solid = None
    solid_results = []
    for cn_result in execute_compute_nodes(context, cn_nodes):
        cn_node = cn_result.compute_node

        if solid and solid is not cn_node.solid:
            yield SolidExecutionResult.from_results(context, solid_results)
            solid_results = []

        if not cn_result.success:
            solid_results.append(cn_result)
            yield SolidExecutionResult.from_results(context, solid_results)
            solid_results = []
            return

        solid = cn_node.solid
        solid_results.append(cn_result)

    if solid and solid_results:
        yield SolidExecutionResult.from_results(context, solid_results)


def execute_pipeline(
    pipeline,
    environment=None,
    throw_on_error=True,
):
    '''
    "Synchronous" version of `execute_pipeline_iterator`.

    Note: throw_on_error is very useful in testing contexts when not testing for error conditions

    Parameters:
      pipeline (PipelineDefinition): pipeline to run
      execution (ExecutionContext): execution context of the run
      throw_on_error (bool):
        throw_on_error makes the function throw when an error is encoutered rather than returning
        the py:class:`SolidExecutionResult` in an error-state.


    Returns:
      PipelineExecutionResult
    '''

    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    environment = check.opt_inst_param(
        environment,
        'environment',
        config.Environment,
        config.Environment(),
    )
    execution_graph = ExecutionGraph.from_pipeline(pipeline)
    return _execute_graph(execution_graph, environment, throw_on_error)


def _execute_graph(
    execution_graph,
    environment,
    throw_on_error=True,
):
    check.inst_param(execution_graph, 'execution_graph', ExecutionGraph)
    check.inst_param(environment, 'environment', config.Environment)
    check.bool_param(throw_on_error, 'throw_on_error')

    display_name = execution_graph.pipeline.display_name
    results = []
    with yield_context(execution_graph.pipeline, environment) as context, \
         context.value('pipeline', execution_graph.pipeline.display_name):

        context.info('Beginning execution of pipeline {pipeline}'.format(pipeline=display_name))

        for result in _execute_graph_iterator(context, execution_graph, environment):
            if throw_on_error and not result.success:
                result.reraise_user_error()

            results.append(result)

        pipeline_result = PipelineExecutionResult(execution_graph.pipeline, context, results)
        if pipeline_result.success:
            context.info(
                'Completing successful execution of pipeline {pipeline}'.format(
                    pipeline=display_name
                )
            )
        else:
            context.info(
                'Completing failing execution of pipeline {pipeline}'.format(pipeline=display_name)
            )

        return pipeline_result



from __future__ import absolute_import
import copy
import json
import logging
from logging import (CRITICAL, DEBUG, ERROR, INFO, WARNING)
import traceback

import coloredlogs

from dagster import check

VALID_LEVELS = set([CRITICAL, DEBUG, ERROR, INFO, WARNING])

LOOKUP = {
    'CRITICAL': CRITICAL,
    'DEBUG': DEBUG,
    'ERROR': ERROR,
    'INFO': INFO,
    'WARNING': WARNING,
}


def level_from_string(string):
    check.str_param(string, 'string')
    return LOOKUP[string]


class CompositeLogger(object):
    def __init__(self, loggers=None):
        self.loggers = check.opt_list_param(loggers, 'loggers', of_type=logging.Logger)

    def __getattr__(self, name):
        def _invoke_logger_method(*args, **kwargs):
            for logger in self.loggers:
                logger_method = check.is_callable(getattr(logger, name))
                logger_method(*args, **kwargs)

        return _invoke_logger_method


class JsonFileHandler(logging.Handler):
    def __init__(self, json_path):
        super(JsonFileHandler, self).__init__()
        self.json_path = check.str_param(json_path, 'json_path')

    def emit(self, record):
        try:
            logged_properties = copy.copy(record.__dict__)
            with open(self.json_path, 'a') as ff:
                text_line = json.dumps(logged_properties)
                ff.write(text_line + '\n')
        # Need to catch Exception here, so disabling lint
        except Exception as e:  # pylint: disable=W0703
            logging.critical('Error during logging!')
            logging.exception(str(e))


def define_json_file_logger(name, json_path, level):
    check.str_param(name, 'name')
    check.str_param(json_path, 'json_path')
    check.param_invariant(
        level in VALID_LEVELS,
        'level',
        'Must be valid python logging level. Got {level}'.format(level=level),
    )

    klass = logging.getLoggerClass()
    logger = klass(name, level=level)
    stream_handler = JsonFileHandler(json_path)
    stream_handler.setFormatter(define_default_formatter())
    logger.addHandler(stream_handler)
    return logger


def define_colored_console_logger(name, level=INFO):
    check.str_param(name, 'name')
    check.param_invariant(
        level in VALID_LEVELS,
        'level',
        'Must be valid python logging level. Got {level}'.format(level=level),
    )

    klass = logging.getLoggerClass()
    logger = klass(name, level=level)
    coloredlogs.install(logger=logger, level=level, fmt=default_format_string())
    return logger


def default_format_string():
    return '%(asctime)s - %(name)s - %(levelname)s - %(message)s'


def define_default_formatter():
    return logging.Formatter(default_format_string())


def debug_format_string():
    return '''%(name)s.%(levelname)s: %(message)s
    time: %(asctime)s relative: %(relativeCreated)dms
    path: %(pathname)s line: %(lineno)d'''


def define_debug_formatter():
    return logging.Formatter(debug_format_string())


def get_formatted_stack_trace(exception):
    check.inst_param(exception, 'exception', Exception)
    if hasattr(exception, '__traceback__'):
        tb = exception.__traceback__
    else:
        import sys
        _exc_type, _exc_value, tb = sys.exc_info()
    return ''.join(traceback.format_tb(tb))



# pylint: disable=W0613

import pytest

import pandas as pd

import dagster.pandas as dagster_pd
from dagster import (
    DagsterInvariantViolationError,
    DagsterTypeError,
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    SolidDefinition,
    config,
    execute_pipeline,
    lambda_solid,
)
from dagster.core.utility_solids import define_stub_solid


def test_wrong_output_value():
    csv_input = InputDefinition('num_csv', dagster_pd.DataFrame)

    @lambda_solid(
        name="test_wrong_output",
        inputs=[csv_input],
        output=OutputDefinition(dagster_pd.DataFrame),
    )
    def df_solid(num_csv):
        return 'not a dataframe'

    pass_solid = define_stub_solid('pass_solid', pd.DataFrame())

    pipeline = PipelineDefinition(
        solids=[pass_solid, df_solid],
        dependencies={'test_wrong_output': {
            'num_csv': DependencyDefinition('pass_solid'),
        }}
    )

    with pytest.raises(DagsterInvariantViolationError):
        execute_pipeline(pipeline)


def test_wrong_input_value():
    @lambda_solid(
        name="test_wrong_input",
        inputs=[InputDefinition('foo', dagster_pd.DataFrame)],
    )
    def df_solid(foo):
        return foo

    pass_solid = define_stub_solid('pass_solid', 'not a dataframe')

    pipeline = PipelineDefinition(
        solids=[pass_solid, df_solid],
        dependencies={'test_wrong_input': {
            'foo': DependencyDefinition('pass_solid'),
        }}
    )

    with pytest.raises(DagsterTypeError):
        execute_pipeline(pipeline)



from collections import namedtuple
import logging
import uuid

from dagster.core.execution import ExecutionContext
from dagster.utils.logging import (DEBUG, INFO, WARNING, ERROR, CRITICAL)

LogMessage = namedtuple('LogMessage', 'msg extra level')

# eliminate complaint about overriding methods imprecisely
# pylint: disable=W0221


# weird name so that pytest ignores
class LoggerForTest(logging.Logger):
    def __init__(self, name=None):
        super(LoggerForTest, self).__init__(name if name else str(uuid.uuid4()))
        self.messages = []

    def debug(self, msg, *args, **kwargs):
        extra = kwargs.pop('extra', None)
        self.messages.append(LogMessage(msg=msg, level=DEBUG, extra=extra))

    def info(self, msg, *args, **kwargs):
        extra = kwargs.pop('extra', None)
        self.messages.append(LogMessage(msg=msg, level=INFO, extra=extra))

    def warning(self, msg, *args, **kwargs):
        extra = kwargs.pop('extra', None)
        self.messages.append(LogMessage(msg=msg, level=WARNING, extra=extra))

    def error(self, msg, *args, **kwargs):
        extra = kwargs.pop('extra', None)
        self.messages.append(LogMessage(msg=msg, level=ERROR, extra=extra))

    def critical(self, msg, *args, **kwargs):
        extra = kwargs.pop('extra', None)
        self.messages.append(LogMessage(msg=msg, level=CRITICAL, extra=extra))


def test_test_logger():
    logger = LoggerForTest()
    logger.debug('fog')
    assert len(logger.messages) == 1
    assert logger.messages[0] == LogMessage(msg='fog', level=DEBUG, extra=None)


def orig_message(message):
    return message.extra['orig_message']


def test_context_logging():
    logger = LoggerForTest()
    context = ExecutionContext(loggers=[logger])
    context.debug('debug from context')
    context.info('info from context')
    context.warning('warning from context')
    context.error('error from context')
    context.critical('critical from context')

    assert len(logger.messages) == 5

    assert logger.messages[0].level == DEBUG
    assert orig_message(logger.messages[0]) == 'debug from context'
    assert logger.messages[1].level == INFO
    assert orig_message(logger.messages[1]) == 'info from context'
    assert logger.messages[2].level == WARNING
    assert orig_message(logger.messages[2]) == 'warning from context'
    assert logger.messages[3].level == ERROR
    assert orig_message(logger.messages[3]) == 'error from context'
    assert logger.messages[4].level == CRITICAL
    assert orig_message(logger.messages[4]) == 'critical from context'


def test_context_value():
    logger = LoggerForTest()
    context = ExecutionContext(loggers=[logger])

    with context.value('some_key', 'some_value'):
        context.info('some message')

    assert logger.messages[0].extra['some_key'] == 'some_value'
    assert 'some_key="some_value"' in logger.messages[0].msg
    assert 'message="some message"' in logger.messages[0].msg


def test_log_message_id():
    logger = LoggerForTest()
    context = ExecutionContext(loggers=[logger])
    context.info('something')

    assert isinstance(uuid.UUID(logger.messages[0].extra['log_message_id']), uuid.UUID)


def test_interleaved_context_value():
    logger = LoggerForTest()
    context = ExecutionContext(loggers=[logger])

    with context.value('key_one', 'value_one'):
        context.info('message one')
        with context.value('key_two', 'value_two'):
            context.info('message two')

    message_one = logger.messages[0]
    assert message_one.extra['key_one'] == 'value_one'
    assert 'key_two' not in message_one.extra
    assert 'key_one="value_one"' in message_one.msg
    assert 'key_two' not in message_one.msg

    message_two = logger.messages[1]
    assert message_two.extra['key_one'] == 'value_one'
    assert message_two.extra['key_two'] == 'value_two'
    assert 'key_one="value_one"' in message_two.msg
    assert 'key_two="value_two"' in message_two.msg


def test_message_specific_logging():
    logger = LoggerForTest()
    context = ExecutionContext(loggers=[logger])
    with context.value('key_one', 'value_one'):
        context.info('message one', key_two='value_two')

    message_one = logger.messages[0]
    assert message_one.extra['key_one'] == 'value_one'
    assert message_one.extra['key_two'] == 'value_two'

    assert set(message_one.extra.keys()) == set(
        ['key_one', 'key_two', 'log_message_id', 'orig_message']
    )


def test_multicontext_value():
    logger = LoggerForTest()
    context = ExecutionContext(loggers=[logger])
    with context.values({
        'key_one': 'value_one',
        'key_two': 'value_two',
    }):
        context.info('message one')

    message_two = logger.messages[0]
    assert message_two.extra['key_one'] == 'value_one'
    assert message_two.extra['key_two'] == 'value_two'
    assert 'key_one="value_one"' in message_two.msg
    assert 'key_two="value_two"' in message_two.msg



from dagster import (DependencyDefinition)


def test_deps_equal():
    assert DependencyDefinition('foo') == DependencyDefinition('foo')
    assert DependencyDefinition('foo') != DependencyDefinition('bar')

    assert DependencyDefinition('foo', 'bar') == DependencyDefinition('foo', 'bar')
    assert DependencyDefinition('foo', 'bar') != DependencyDefinition('foo', 'quuz')



from __future__ import print_function
from collections import (namedtuple, defaultdict)
from contextlib import contextmanager
from enum import Enum
import uuid
import sys

from future.utils import raise_from

import toposort

from dagster import (
    check,
    config,
)

from dagster.utils.indenting_printer import IndentingPrinter
from dagster.utils.logging import (
    get_formatted_stack_trace,
    define_colored_console_logger,
    ERROR,
)

from dagster.utils.timing import time_execution_scope

from .definitions import (
    ExecutionGraph,
    ExpectationDefinition,
    ExpectationExecutionInfo,
    InputDefinition,
    OutputDefinition,
    Result,
    SolidDefinition,
    SolidOutputHandle,
)

from .execution_context import (
    ExecutionContext,
    TransformExecutionInfo,
)

from .errors import (
    DagsterError,
    DagsterExpectationFailedError,
    DagsterInvariantViolationError,
    DagsterTypeError,
    DagsterUserCodeExecutionError,
    DagsterEvaluateValueError,
)

from .types import DagsterType


class ComputeNodeExecutionInfo(
    namedtuple('_ComputeNodeExecutionInfo', 'context execution_graph environment')
):
    def __new__(cls, context, execution_graph, environment):
        return super(ComputeNodeExecutionInfo, cls).__new__(
            cls,
            check.inst_param(context, 'context', ExecutionContext),
            check.inst_param(execution_graph, 'execution_graph', ExecutionGraph),
            check.inst_param(environment, 'environment', config.Environment),
        )

    @property
    def pipeline(self):
        return self.execution_graph.pipeline


class ComputeNodeOutputHandle(namedtuple('_ComputeNodeOutputHandle', 'compute_node output_name')):
    def __new__(cls, compute_node, output_name):
        return super(ComputeNodeOutputHandle, cls).__new__(
            cls,
            compute_node=check.inst_param(compute_node, 'compute_node', ComputeNode),
            output_name=check.str_param(output_name, 'output_name'),
        )

    # Make this hashable so it be a key in a dictionary

    def __str__(self):
        return (
            'ComputeNodeOutputHandle'
            '(cn="{cn.compute_node.friendly_name}", output_name="{cn.output_name}")'.format(
                cn=self
            )
        )

    def __repr__(self):
        return (
            'ComputeNodeOutputHandle'
            '(cn="{cn.compute_node.friendly_name}", output_name="{cn.output_name}")'.format(
                cn=self
            )
        )

    def __hash__(self):
        return hash(self.compute_node.guid + self.output_name)

    def __eq__(self, other):
        return (
            self.compute_node.guid == other.compute_node.guid
            and self.output_name == other.output_name
        )


class ComputeNodeSuccessData(namedtuple('_ComputeNodeSuccessData', 'output_name value')):
    def __new__(cls, output_name, value):
        return super(ComputeNodeSuccessData, cls).__new__(
            cls,
            output_name=check.str_param(output_name, 'output_name'),
            value=value,
        )


class ComputeNodeFailureData(namedtuple('_ComputeNodeFailureData', 'dagster_error')):
    def __new__(cls, dagster_error):
        return super(ComputeNodeFailureData, cls).__new__(
            cls,
            dagster_error=check.inst_param(dagster_error, 'dagster_error', DagsterError),
        )


class ComputeNodeResult(
    namedtuple(
        '_ComputeNodeResult',
        'success compute_node tag success_data failure_data',
    )
):
    @staticmethod
    def success_result(compute_node, tag, success_data):
        return ComputeNodeResult(
            success=True,
            compute_node=check.inst_param(compute_node, 'compute_node', ComputeNode),
            tag=check.inst_param(tag, 'tag', ComputeNodeTag),
            success_data=check.inst_param(success_data, 'success_data', ComputeNodeSuccessData),
            failure_data=None,
        )

    @staticmethod
    def failure_result(compute_node, tag, failure_data):
        return ComputeNodeResult(
            success=False,
            compute_node=check.inst_param(compute_node, 'compute_node', ComputeNode),
            tag=check.inst_param(tag, 'tag', ComputeNodeTag),
            success_data=None,
            failure_data=check.inst_param(failure_data, 'failure_data', ComputeNodeFailureData),
        )


@contextmanager
def _user_code_error_boundary(context, msg, **kwargs):
    '''
    Wraps the execution of user-space code in an error boundary. This places a uniform
    policy around an user code invoked by the framework. This ensures that all user
    errors are wrapped in the SolidUserCodeExecutionError, and that the original stack
    trace of the user error is preserved, so that it can be reported without confusing
    framework code in the stack trace, if a tool author wishes to do so. This has
    been especially help in a notebooking context.
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.str_param(msg, 'msg')

    try:
        yield
    except DagsterError as de:
        stack_trace = get_formatted_stack_trace(de)
        context.error(str(de), stack_trace=stack_trace)
        raise de
    except Exception as e: # pylint: disable=W0703
        stack_trace = get_formatted_stack_trace(e)
        context.error(str(e), stack_trace=stack_trace)
        raise_from(
            DagsterUserCodeExecutionError(
                msg.format(**kwargs),
                user_exception=e,
                original_exc_info=sys.exc_info(),
            ),
            e,
        )


class ComputeNodeTag(Enum):
    TRANSFORM = 'TRANSFORM'
    INPUT_EXPECTATION = 'INPUT_EXPECTATION'
    OUTPUT_EXPECTATION = 'OUTPUT_EXPECTATION'
    JOIN = 'JOIN'


EXPECTATION_VALUE_OUTPUT = 'expectation_value'
JOIN_OUTPUT = 'join_output'
EXPECTATION_INPUT = 'expectation_input'



def _yield_transform_results(context, compute_node, conf, inputs):
    gen = compute_node.solid.transform_fn(TransformExecutionInfo(context, conf), inputs)

    if isinstance(gen, Result):
        raise DagsterInvariantViolationError(
            (
                'Transform for solid {solid_name} returned a Result rather than ' +
                'yielding it. The transform_fn of the core SolidDefinition must yield ' +
                'its results'
            ).format(solid_name=compute_node.solid.name)
        )

    if gen is None:
        return

    for result in gen:
        if not isinstance(result, Result):
            raise DagsterInvariantViolationError(
                (
                    'Transform for solid {solid_name} yielded {result} rather an ' +
                    'an instance of the Result class.'
                ).format(
                    result=repr(result),
                    solid_name=compute_node.solid.name,
                )
            )
        yield result


def _collect_result_list(context, compute_node, conf, inputs):
    for result in _yield_transform_results(context, compute_node, conf, inputs):
        context.info(
            'Solid {solid} emitted output "{output}" value {value}'.format(
                solid=compute_node.solid.name,
                output=result.output_name,
                value=repr(result.value),
            )
        )
        yield result


def _execute_core_transform(context, compute_node, conf, inputs):
    '''
    Execute the user-specified transform for the solid. Wrap in an error boundary and do
    all relevant logging and metrics tracking
    '''
    check.inst_param(context, 'context', ExecutionContext)
    check.inst_param(compute_node, 'compute_node', ComputeNode)
    check.dict_param(inputs, 'inputs', key_type=str)

    error_str = 'Error occured during core transform'

    solid = compute_node.solid

    with context.value('solid', solid.name):
        context.debug(
            'Executing core transform for solid {solid}.'.format(solid=solid.name)
        )

        with time_execution_scope() as timer_result, \
            _user_code_error_boundary(context, error_str):

            all_results = list(_collect_result_list(context, compute_node, conf, inputs))

        if len(all_results) != len(solid.output_defs):
            emitted_result_names = set([r.output_name for r in all_results])
            solid_output_names = set([output_def.name for output_def in solid.output_defs])
            omitted_outputs = solid_output_names.difference(emitted_result_names)
            context.info('Solid {solid} did not fire outputs {outputs}'.format(
                solid=solid.name,
                outputs=repr(omitted_outputs),
            ))

        context.debug(
            'Finished executing transform for solid {solid}. Time elapsed: {millis:.3f} ms'.format(
                solid=compute_node.solid.name,
                millis=timer_result.millis,
            ),
            execution_time_ms=timer_result.millis,
        )

        for result in all_results:
            yield result


class ComputeNodeInput(object):
    def __init__(self, name, dagster_type, prev_output_handle):
        self.name = check.str_param(name, 'name')
        self.dagster_type = check.inst_param(dagster_type, 'dagster_type', DagsterType)
        self.prev_output_handle = check.inst_param(
            prev_output_handle,
            'prev_output_handle',
            ComputeNodeOutputHandle,
        )


class ComputeNodeOutput(object):
    def __init__(self, name, dagster_type):
        self.name = check.str_param(name, 'name')
        self.dagster_type = check.inst_param(dagster_type, 'dagster_type', DagsterType)


class ComputeNode(object):
    def __init__(self, friendly_name, node_inputs, node_outputs, arg_dict, compute_fn, tag, solid):
        self.guid = str(uuid.uuid4())
        self.friendly_name = check.str_param(friendly_name, 'friendly_name')
        self.node_inputs = check.list_param(node_inputs, 'node_inputs', of_type=ComputeNodeInput)

        node_input_dict = {}
        for node_input in node_inputs:
            node_input_dict[node_input.name] = node_input
        self._node_input_dict = node_input_dict
        self.node_outputs = check.list_param(
            node_outputs, 'node_outputs', of_type=ComputeNodeOutput
        )

        node_output_dict = {}
        for node_output in node_outputs:
            node_output_dict[node_output.name] = node_output

        self._node_output_dict = node_output_dict
        self.arg_dict = check.dict_param(arg_dict, 'arg_dict', key_type=str)
        self.compute_fn = check.callable_param(compute_fn, 'compute_fn')
        self.tag = check.inst_param(tag, 'tag', ComputeNodeTag)
        self.solid = check.inst_param(solid, 'solid', SolidDefinition)

    def has_node(self, name):
        check.str_param(name, 'name')
        return name in self._node_output_dict

    def node_named(self, name):
        check.str_param(name, 'name')
        return self._node_output_dict[name]

    def _create_compute_node_result(self, result):
        check.inst_param(result, 'result', Result)

        node_output = self.node_named(result.output_name)

        try:
            evaluation_result = node_output.dagster_type.evaluate_value(result.value)
        except DagsterEvaluateValueError as e:
            raise DagsterInvariantViolationError(
                '''Solid {cn.solid.name} output name {output_name} output {result.value}
                type failure: {error_msg}'''.format(
                    cn=self,
                    result=result,
                    error_msg=','.join(e.args),
                    output_name=result.output_name,
                )
            )

        return ComputeNodeResult.success_result(
            compute_node=self,
            tag=self.tag,
            success_data=ComputeNodeSuccessData(
                output_name=result.output_name,
                value=evaluation_result,
            ),
        )

    def _get_evaluated_input(self, input_name, input_value):
        compute_node_input = self._node_input_dict[input_name]
        try:
            return compute_node_input.dagster_type.evaluate_value(input_value)
        except DagsterEvaluateValueError as evaluate_error:
            raise_from(
                DagsterTypeError((
                    'Solid {cn.solid.name} input {input_name} received value {input_value} ' +
                    'which does not pass the typecheck for Dagster type ' +
                    '{compute_node_input.dagster_type.name}. Compute node {cn.friendly_name}'
                    ).format(
                        cn=self,
                        input_name=input_name,
                        input_value=input_value,
                        compute_node_input=compute_node_input,
                    )
                ),
                evaluate_error,
            )

    def _compute_result_list(self, context, evaluated_inputs):
        error_str = 'Error occured during compute node {friendly_name}'.format(
            friendly_name=self.friendly_name,
        )

        with _user_code_error_boundary(context, error_str):
            gen = self.compute_fn(context, self, evaluated_inputs)

            if gen is None:
                check.invariant(not self.node_outputs)
                return

            return list(gen)

    def _error_check_results(self, results):
        seen_outputs = set()
        for result in results:
            if not self.has_node(result.output_name):
                output_names = list([output_def.name for output_def in self.solid.output_defs])
                raise DagsterInvariantViolationError(
                    '''Core transform for {cn.solid.name} returned an output
                    {result.output_name} that does not exist. The available
                    outputs are {output_names}'''
                    .format(cn=self, result=result, output_names=output_names)
                )

            if result.output_name in seen_outputs:
                raise DagsterInvariantViolationError(
                    '''Core transform for {cn.solid.name} returned an output
                    {result.output_name} multiple times'''.format(cn=self, result=result)
                )

            seen_outputs.add(result.output_name)


    def _execute_inner_compute_node_loop(self, context, inputs):
        evaluated_inputs = {}
        # do runtime type checks of inputs versus node inputs
        for input_name, input_value in inputs.items():
            evaluated_inputs[input_name] = self._get_evaluated_input(input_name, input_value)

        results = self._compute_result_list(context, evaluated_inputs)

        self._error_check_results(results)

        return [self._create_compute_node_result(result) for result in results]

    def execute(self, context, inputs):
        check.inst_param(context, 'context', ExecutionContext)
        check.dict_param(inputs, 'inputs', key_type=str)

        try:
            for compute_node_result in self._execute_inner_compute_node_loop(context, inputs):
                yield compute_node_result

        except DagsterError as dagster_error:
            context.error(str(dagster_error))
            yield ComputeNodeResult.failure_result(
                compute_node=self,
                tag=self.tag,
                failure_data=ComputeNodeFailureData(
                    dagster_error=dagster_error,
                ),
            )
            return

    def output_named(self, name):
        check.str_param(name, 'name')

        for node_output in self.node_outputs:
            if node_output.name == name:
                return node_output

        check.failed('output {name} not found'.format(name=name))


def _all_inputs_covered(cn, results):
    for node_input in cn.node_inputs:
        if node_input.prev_output_handle not in results:
            return False
    return True


def execute_compute_nodes(context, compute_nodes):
    check.inst_param(context, 'context', ExecutionContext)
    check.list_param(compute_nodes, 'compute_nodes', of_type=ComputeNode)

    intermediate_results = {}
    context.debug(
        'Entering execute_compute_nodes loop. Order: {order}'.format(
            order=[cn.friendly_name for cn in compute_nodes]
        )
    )

    for compute_node in compute_nodes:
        if not _all_inputs_covered(compute_node, intermediate_results):
            result_keys = set(intermediate_results.keys())
            expected_outputs = [ni.prev_output_handle for ni in compute_node.node_inputs]

            context.debug(
                'Not all inputs covered for {compute_name}. Not executing.'.
                format(compute_name=compute_node.friendly_name) +
                '\nKeys in result: {result_keys}.'.format(result_keys=result_keys, ) +
                '\nOutputs need for inputs {expected_outputs}'.
                format(expected_outputs=expected_outputs, )
            )
            continue

        input_values = {}
        for node_input in compute_node.node_inputs:
            prev_output_handle = node_input.prev_output_handle
            input_value = intermediate_results[prev_output_handle].success_data.value
            input_values[node_input.name] = input_value

        for result in compute_node.execute(context, input_values):
            check.invariant(isinstance(result, ComputeNodeResult))
            yield result
            output_handle = ComputeNodeOutputHandle(compute_node, result.success_data.output_name)
            intermediate_results[output_handle] = result


def print_graph(graph, printer=print):
    check.inst_param(graph, 'graph', ComputeNodeGraph)
    printer = IndentingPrinter(printer=printer)

    for node in graph.topological_nodes():
        with printer.with_indent('Node {node.friendly_name} Id: {node.guid}'.format(node=node)):
            for node_input in node.node_inputs:
                with printer.with_indent('Input: {node_input.name}'.format(node_input=node_input)):
                    printer.line(
                        'Type: {node_input.dagster_type.name}'.format(node_input=node_input)
                    )
                    printer.line(
                        'From: {node_input.prev_output_handle}'.format(node_input=node_input)
                    )
            for node_output in node.node_outputs:
                with printer.with_indent(
                    'Output: {node_output.name}'.format(node_output=node_output)
                ):
                    printer.line(
                        'Type: {node_output.dagster_type.name}'.format(node_output=node_output)
                    )


class ComputeNodeGraph(object):
    def __init__(self, cn_dict, deps):
        self.cn_dict = cn_dict
        self.deps = deps
        self.nodes = list(cn_dict.values())

    def topological_nodes(self):
        cn_guids_sorted = toposort.toposort_flatten(self.deps)
        for cn_guid in cn_guids_sorted:
            yield self.cn_dict[cn_guid]


def create_expectation_cn(
    solid,
    expectation_def,
    friendly_name,
    tag,
    prev_node_output_handle,
    inout_def,
):

    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(expectation_def, 'input_expct_def', ExpectationDefinition)
    check.inst_param(prev_node_output_handle, 'prev_node_output_handle', ComputeNodeOutputHandle)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))

    value_type = inout_def.dagster_type

    return ComputeNode(
        friendly_name=friendly_name,
        node_inputs=[
            ComputeNodeInput(
                name=EXPECTATION_INPUT,
                dagster_type=value_type,
                prev_output_handle=prev_node_output_handle,
            )
        ],
        node_outputs=[
            ComputeNodeOutput(name=EXPECTATION_VALUE_OUTPUT, dagster_type=value_type),
        ],
        arg_dict={},
        compute_fn=_create_expectation_lambda(
            solid,
            inout_def,
            expectation_def,
            EXPECTATION_VALUE_OUTPUT,
        ),
        tag=tag,
        solid=solid
    )


ComputeNodeSubgraph = namedtuple(
    'ComputeNodeSubgraph',
    'nodes terminal_cn_output_handle',
)


def create_expectations_cn_graph(solid, inout_def, prev_node_output_handle, tag):
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(prev_node_output_handle, 'prev_node_output_handle', ComputeNodeOutputHandle)
    check.inst_param(tag, 'tag', ComputeNodeTag)

    compute_nodes = []
    input_expect_nodes = []
    for expectation_def in inout_def.expectations:
        expect_compute_node = create_expectation_cn(
            solid=solid,
            expectation_def=expectation_def,
            friendly_name='{solid.name}.{inout_def.name}.expectation.{expectation_def.name}'.format(
                solid=solid, inout_def=inout_def, expectation_def=expectation_def
            ),
            tag=tag,
            prev_node_output_handle=prev_node_output_handle,
            inout_def=inout_def,
        )
        input_expect_nodes.append(expect_compute_node)
        compute_nodes.append(expect_compute_node)

    join_cn = _create_join_node(solid, input_expect_nodes, EXPECTATION_VALUE_OUTPUT)

    output_name = join_cn.node_outputs[0].name
    return ComputeNodeSubgraph(
        compute_nodes + [join_cn],
        ComputeNodeOutputHandle(join_cn, output_name),
    )


class ComputeNodeOutputMap(dict):
    def __getitem__(self, key):
        check.inst_param(key, 'key', SolidOutputHandle)
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        check.inst_param(key, 'key', SolidOutputHandle)
        check.inst_param(val, 'val', ComputeNodeOutputHandle)
        return dict.__setitem__(self, key, val)


def create_conf_value(execution_info, solid):
    check.inst_param(execution_info, 'execution_info', ComputeNodeExecutionInfo)
    check.inst_param(solid, 'solid', SolidDefinition)

    name = solid.name
    solid_configs = execution_info.environment.solids
    config_input = solid_configs[name].config if name in solid_configs else {}

    try:
        return solid.config_def.config_type.evaluate_value(config_input)
    except DagsterEvaluateValueError as e:
        raise DagsterTypeError(
            'Error evaluating config for {solid_name}: {error_msg}'.format(
                solid_name=solid.name,
                error_msg=','.join(e.args),
            )
        )


def create_compute_node_graph(execution_info):
    check.inst_param(execution_info, 'execution_info', ComputeNodeExecutionInfo)

    execution_graph = execution_info.execution_graph

    dependency_structure = execution_graph.dependency_structure

    compute_nodes = []

    cn_output_node_map = ComputeNodeOutputMap()

    for topo_solid in execution_graph.topological_solids:
        cn_inputs = []

        for input_def in topo_solid.input_defs:
            input_handle = topo_solid.input_handle(input_def.name)

            check.invariant(
                dependency_structure.has_dep(input_handle),
                '{input_handle} not found in dependency structure'.format(
                    input_handle=input_handle
                ),
            )

            solid_output_handle = dependency_structure.get_dep(input_handle)
            prev_cn_output_handle = cn_output_node_map[solid_output_handle]

            subgraph = create_subgraph_for_input(
                execution_info,
                topo_solid,
                prev_cn_output_handle,
                input_def,
            )

            compute_nodes.extend(subgraph.nodes)
            cn_inputs.append(
                ComputeNodeInput(
                    input_def.name,
                    input_def.dagster_type,
                    subgraph.terminal_cn_output_handle,
                )
            )

        conf = create_conf_value(execution_info, topo_solid)

        solid_transform_cn = create_compute_node_from_solid_transform(topo_solid, cn_inputs, conf)

        compute_nodes.append(solid_transform_cn)

        for output_def in topo_solid.output_defs:
            subgraph = create_subgraph_for_output(
                execution_info,
                topo_solid,
                solid_transform_cn,
                output_def,
            )
            compute_nodes.extend(subgraph.nodes)

            output_handle = topo_solid.output_handle(output_def.name)
            cn_output_node_map[output_handle] = subgraph.terminal_cn_output_handle

    return _create_compute_node_graph(compute_nodes)


def _create_compute_node_graph(compute_nodes):
    cn_dict = {}
    for cn in compute_nodes:
        cn_dict[cn.guid] = cn

    deps = defaultdict()

    for cn in compute_nodes:
        deps[cn.guid] = set()
        for cn_input in cn.node_inputs:
            deps[cn.guid].add(cn_input.prev_output_handle.compute_node.guid)

    return ComputeNodeGraph(cn_dict, deps)


def create_subgraph_for_input(execution_info, solid, prev_cn_output_handle, input_def):
    check.inst_param(execution_info, 'execution_info', ComputeNodeExecutionInfo)
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(prev_cn_output_handle, 'prev_cn_output_handle', ComputeNodeOutputHandle)
    check.inst_param(input_def, 'input_def', InputDefinition)

    if execution_info.environment.expectations.evaluate and input_def.expectations:
        return create_expectations_cn_graph(
            solid,
            input_def,
            prev_cn_output_handle,
            tag=ComputeNodeTag.INPUT_EXPECTATION,
        )
    else:
        return ComputeNodeSubgraph(
            nodes=[],
            terminal_cn_output_handle=prev_cn_output_handle,
        )


def create_subgraph_for_output(execution_info, solid, solid_transform_cn, output_def):
    check.inst_param(execution_info, 'execution_info', ComputeNodeExecutionInfo)
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(solid_transform_cn, 'solid_transform_cn', ComputeNode)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    if execution_info.environment.expectations.evaluate and output_def.expectations:
        return create_expectations_cn_graph(
            solid,
            output_def,
            ComputeNodeOutputHandle(solid_transform_cn, output_def.name),
            tag=ComputeNodeTag.OUTPUT_EXPECTATION
        )
    else:
        return ComputeNodeSubgraph(
            nodes=[],
            terminal_cn_output_handle=ComputeNodeOutputHandle(
                solid_transform_cn,
                output_def.name,
            ),
        )


def _create_join_node(solid, prev_nodes, prev_output_name):
    check.inst_param(solid, 'solid', SolidDefinition)
    check.list_param(prev_nodes, 'prev_nodes', of_type=ComputeNode)
    check.invariant(len(prev_nodes) > 0)
    check.str_param(prev_output_name, 'output_name')

    node_inputs = []
    seen_dagster_type = None
    for prev_node in prev_nodes:
        prev_node_output = prev_node.output_named(prev_output_name)

        if seen_dagster_type is None:
            seen_dagster_type = prev_node_output.dagster_type
        else:
            check.invariant(seen_dagster_type == prev_node_output.dagster_type)

        output_handle = ComputeNodeOutputHandle(prev_node, prev_output_name)

        node_inputs.append(
            ComputeNodeInput(prev_node.guid, prev_node_output.dagster_type, output_handle)
        )

    return ComputeNode(
        friendly_name='join',
        node_inputs=node_inputs,
        node_outputs=[ComputeNodeOutput(JOIN_OUTPUT, seen_dagster_type)],
        arg_dict={},
        compute_fn=_create_join_lambda,
        tag=ComputeNodeTag.JOIN,
        solid=solid,
    )


def _create_join_lambda(_context, _compute_node, inputs):
    yield Result(output_name=JOIN_OUTPUT, value=list(inputs.values())[0])


def _create_expectation_lambda(solid, inout_def, expectation_def, internal_output_name):
    check.inst_param(solid, 'solid', SolidDefinition)
    check.inst_param(inout_def, 'inout_def', (InputDefinition, OutputDefinition))
    check.inst_param(expectation_def, 'expectations_def', ExpectationDefinition)
    check.str_param(internal_output_name, 'internal_output_name')

    def _do_expectation(context, compute_node, inputs):
        with context.values({
            'solid': solid.name,
            inout_def.descriptive_key: inout_def.name,
            'expectation': expectation_def.name
        }):
            value = inputs[EXPECTATION_INPUT]
            info = ExpectationExecutionInfo(context, inout_def, solid, expectation_def)
            expt_result = expectation_def.expectation_fn(info, value)
            if expt_result.success:
                context.debug('Expectation {friendly_name} succeeded on {value}.'.format(
                    friendly_name=compute_node.friendly_name,
                    value=value,
                ))
                yield Result(output_name=internal_output_name, value=inputs[EXPECTATION_INPUT])
            else:
                context.debug('Expectation {friendly_name} failed on {value}.'.format(
                    friendly_name=compute_node.friendly_name,
                    value=value,
                ))
                raise DagsterExpectationFailedError(info, value)

    return _do_expectation


def create_compute_node_from_solid_transform(solid, node_inputs, conf):
    check.inst_param(solid, 'solid', SolidDefinition)
    check.list_param(node_inputs, 'node_inputs', of_type=ComputeNodeInput)

    return ComputeNode(
        friendly_name='{solid.name}.transform'.format(solid=solid),
        node_inputs=node_inputs,
        node_outputs=[
            ComputeNodeOutput(name=output_def.name, dagster_type=output_def.dagster_type)
            for output_def in solid.output_defs
        ],
        arg_dict={},
        compute_fn=lambda context, compute_node, inputs: _execute_core_transform(
            context,
            compute_node,
            conf,
            inputs,
        ),
        tag=ComputeNodeTag.TRANSFORM,
        solid=solid,
    )



import dagster
from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    lambda_solid,
)
import dagster.pandas as dagster_pd


@lambda_solid(
    inputs=[InputDefinition('num', dagster_pd.DataFrame)],
    output=OutputDefinition(dagster_pd.DataFrame),
)
def sum_solid(num):
    sum_df = num.copy()
    sum_df['sum'] = sum_df['num1'] + sum_df['num2']
    return sum_df


@lambda_solid(
    inputs=[InputDefinition('sum_df', dagster_pd.DataFrame)],
    output=OutputDefinition(dagster_pd.DataFrame),
)
def sum_sq_solid(sum_df):
    sum_sq_df = sum_df.copy()
    sum_sq_df['sum_sq'] = sum_df['sum']**2
    return sum_sq_df


@lambda_solid(
    inputs=[InputDefinition('sum_sq_solid', dagster_pd.DataFrame)],
    output=OutputDefinition(dagster_pd.DataFrame),
)
def always_fails_solid(**_kwargs):
    raise Exception('I am a programmer and I make error')


def define_failure_pipeline():
    return dagster.PipelineDefinition(
        name='pandas_hello_world_fails',
        solids=[
            dagster_pd.load_csv_solid('load_num_csv'),
            sum_solid,
            sum_sq_solid,
            always_fails_solid,
        ],
        dependencies={
            'sum_solid': {
                'num': DependencyDefinition('load_num_csv')
            },
            'sum_sq_solid': {
                'sum_df': DependencyDefinition(sum_solid.name),
            },
            'always_fails_solid': {
                'sum_sq_solid': DependencyDefinition(sum_sq_solid.name),
            }
        }
    )


def define_success_pipeline():
    return PipelineDefinition(
        name='pandas_hello_world',
        solids=[dagster_pd.load_csv_solid('load_num_csv'), sum_solid, sum_sq_solid],
        dependencies={
            'sum_solid': {
                'num': DependencyDefinition('load_num_csv')
            },
            'sum_sq_solid': {
                'sum_df': DependencyDefinition(sum_solid.name),
            },
        },
    )



from dagster import (
    DependencyDefinition,
    InputDefinition,
    OutputDefinition,
    PipelineDefinition,
    check,
    config,
    execute_pipeline,
)

from dagster.core.definitions import (
    DependencyStructure,
    ExecutionGraph,
    _create_adjacency_lists,
)

from dagster.core.execution import (
    execute_pipeline_iterator,
    ExecutionContext,
    SolidExecutionResult,
    PipelineExecutionResult,
)

from dagster.core.test_utils import single_output_transform

from dagster.core.utility_solids import define_stub_solid

# protected members
# pylint: disable=W0212


def _default_passthrough_transform(*args, **kwargs):
    check.invariant(not args, 'There should be no positional args')
    return list(kwargs.values())[0]


def create_dep_input_fn(name):
    return lambda context, arg_dict: {name: 'input_set'}


def make_transform(name):
    def transform(_context, inputs):
        passed_rows = []
        seen = set()
        for row in inputs.values():
            for item in row:
                key = list(item.keys())[0]
                if key not in seen:
                    seen.add(key)
                    passed_rows.append(item)

        result = []
        result.extend(passed_rows)
        result.append({name: 'transform_called'})
        return result

    return transform


def create_solid_with_deps(name, *solid_deps):
    inputs = [InputDefinition(solid_dep.name) for solid_dep in solid_deps]

    return single_output_transform(
        name=name,
        inputs=inputs,
        transform_fn=make_transform(name),
        output=OutputDefinition(),
    )


def create_root_solid(name):
    input_name = name + '_input'
    inp = InputDefinition(input_name)

    return single_output_transform(
        name=name,
        inputs=[inp],
        transform_fn=make_transform(name),
        output=OutputDefinition(),
    )


def _do_construct(solids, dependencies):
    dependency_structure = DependencyStructure.from_definitions(solids, dependencies)
    return _create_adjacency_lists(solids, dependency_structure)


def test_empty_adjaceny_lists():
    solids = [create_root_solid('a_node')]
    forward_edges, backwards_edges = _do_construct(solids, {})
    assert forward_edges == {'a_node': set()}
    assert backwards_edges == {'a_node': set()}


def test_single_dep_adjacency_lists():
    # A <-- B
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)

    forward_edges, backwards_edges = _do_construct(
        [node_a, node_b],
        {
            'B': {
                'A': DependencyDefinition('A'),
            },
        },
    )

    assert forward_edges == {'A': {'B'}, 'B': set()}
    assert backwards_edges == {'B': {'A'}, 'A': set()}


def graph_from_solids_only(solids, dependencies):
    pipeline = PipelineDefinition(solids=solids, dependencies=dependencies)
    return ExecutionGraph.from_pipeline(pipeline)


def test_diamond_deps_adjaceny_lists():
    forward_edges, backwards_edges = _do_construct(
        create_diamond_solids(),
        diamond_deps(),
    )

    assert forward_edges == {'A_source': {'A'}, 'A': {'B', 'C'}, 'B': {'D'}, 'C': {'D'}, 'D': set()}
    assert backwards_edges == {
        'D': {'B', 'C'},
        'B': {'A'},
        'C': {'A'},
        'A': {'A_source'},
        'A_source': set()
    }


def diamond_deps():
    return {
        'A': {
            'A_input': DependencyDefinition('A_source'),
        },
        'B': {
            'A': DependencyDefinition('A')
        },
        'C': {
            'A': DependencyDefinition('A')
        },
        'D': {
            'B': DependencyDefinition('B'),
            'C': DependencyDefinition('C'),
        }
    }


def test_disconnected_graphs_adjaceny_lists():
    # A <-- B
    # C <-- D
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)

    node_c = create_root_solid('C')
    node_d = create_solid_with_deps('D', node_c)

    forward_edges, backwards_edges = _do_construct(
        [node_a, node_b, node_c, node_d], {
            'B': {
                'A': DependencyDefinition('A')
            },
            'D': {
                'C': DependencyDefinition('C'),
            }
        }
    )
    assert forward_edges == {'A': {'B'}, 'B': set(), 'C': {'D'}, 'D': set()}
    assert backwards_edges == {'B': {'A'}, 'A': set(), 'D': {'C'}, 'C': set()}


def create_diamond_solids():
    a_source = define_stub_solid('A_source', [input_set('A_input')])
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)
    node_c = create_solid_with_deps('C', node_a)
    node_d = create_solid_with_deps('D', node_b, node_c)
    return [node_d, node_c, node_b, node_a, a_source]


def create_diamond_graph():
    return graph_from_solids_only(create_diamond_solids(), diamond_deps())


def test_diamond_toposort():
    graph = create_diamond_graph()
    assert graph.topological_order == ['A_source', 'A', 'B', 'C', 'D']


def test_execution_graph_diamond():
    solid_graph = create_diamond_graph()

    full_execution_graph = solid_graph.create_execution_subgraph(
        from_solids=['A'],
        to_solids=['D'],
    )
    assert full_execution_graph.topological_order == ['A', 'B', 'C', 'D']
    assert len(full_execution_graph._solid_dict) == 4

    d_only_graph = solid_graph.create_execution_subgraph(
        from_solids=['D'],
        to_solids=['D'],
    )
    assert len(d_only_graph._solid_dict) == 1
    assert d_only_graph.topological_order == ['D']

    abc_graph = solid_graph.create_execution_subgraph(
        from_solids=['B', 'C'],
        to_solids=['D'],
    )
    assert len(abc_graph._solid_dict) == 3
    assert abc_graph.topological_order == ['B', 'C', 'D']


def input_set(name):
    return {name: 'input_set'}


def transform_called(name):
    return {name: 'transform_called'}


def assert_equivalent_results(left, right):
    check.inst_param(left, 'left', SolidExecutionResult)
    check.inst_param(right, 'right', SolidExecutionResult)

    assert left.success == right.success
    assert left.name == right.name
    assert left.solid.name == right.solid.name
    assert left.transformed_value == right.transformed_value


def assert_all_results_equivalent(expected_results, result_results):
    check.list_param(expected_results, 'expected_results', of_type=SolidExecutionResult)
    check.list_param(result_results, 'result_results', of_type=SolidExecutionResult)
    assert len(expected_results) == len(result_results)
    for expected, result in zip(expected_results, result_results):
        assert_equivalent_results(expected, result)


def test_pipeline_execution_graph_diamond():
    pipeline = PipelineDefinition(solids=create_diamond_solids(), dependencies=diamond_deps())
    environment = config.Environment()
    return _do_test(pipeline, lambda: execute_pipeline_iterator(
        pipeline,
        environment=environment,
    ))


def test_create_single_solid_pipeline():
    stub_solid = define_stub_solid('stub', [{'a key': 'a value'}])
    single_solid_pipeline = PipelineDefinition.create_single_solid_pipeline(
        PipelineDefinition(solids=create_diamond_solids(), dependencies=diamond_deps()),
        'A',
        {
            'A': {
                'A_input': stub_solid
            },
        },
    )

    result = execute_pipeline(single_solid_pipeline)
    assert result.success


# def test_pipeline_execution_graph_diamond_in_memory():
#     pipeline = PipelineDefinition(solids=create_diamond_solids(), dependencies=diamond_deps())
#     # input_values = {'A_input': [{'A_input': 'input_set'}]}
#     return _do_test(lambda: execute_pipeline_iterator_in_memory(
#         ExecutionContext(),
#         pipeline,
#         {}
#         # input_values={'A': input_values},
#     ))


def _do_test(pipeline, do_execute_pipeline_iter):

    results = list()

    for result in do_execute_pipeline_iter():
        results.append(result)

    result = PipelineExecutionResult(pipeline, ExecutionContext(), results)

    assert result.result_for_solid('A').transformed_value() == [
        input_set('A_input'), transform_called('A')
    ]

    assert result.result_for_solid('B').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('B'),
    ]

    assert result.result_for_solid('C').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('C'),
    ]

    assert result.result_for_solid('D').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('C'),
        transform_called('B'),
        transform_called('D'),
    ] or result.result_for_solid('D').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('B'),
        transform_called('C'),
        transform_called('D'),
    ]


def test_empty_pipeline_execution():
    result = execute_pipeline(PipelineDefinition(solids=[]))

    assert result.success



from collections import defaultdict

from dagster import (
    PipelineDefinition,
    RepositoryDefinition,
    SolidDefinition,
)


def create_single_node_pipeline(name, called):
    called[name] = called[name] + 1
    return PipelineDefinition(
        name=name,
        solids=[
            SolidDefinition(
                name=name + '_solid',
                inputs=[],
                outputs=[],
                transform_fn=lambda *_args, **_kwargs: None,
            )
        ]
    )


def test_repo_definition():
    called = defaultdict(int)
    repo = RepositoryDefinition(
        name='some_repo',
        pipeline_dict={
            'foo': lambda: create_single_node_pipeline('foo', called),
            'bar': lambda: create_single_node_pipeline('bar', called),
        }
    )

    foo_pipeline = repo.get_pipeline('foo')
    assert isinstance(foo_pipeline, PipelineDefinition)
    assert foo_pipeline.name == 'foo'

    assert 'foo' in called
    assert called['foo'] == 1
    assert 'bar' not in called

    bar_pipeline = repo.get_pipeline('bar')
    assert isinstance(bar_pipeline, PipelineDefinition)
    assert bar_pipeline.name == 'bar'

    assert 'foo' in called
    assert called['foo'] == 1
    assert 'bar' in called
    assert called['bar'] == 1

    foo_pipeline = repo.get_pipeline('foo')
    assert isinstance(foo_pipeline, PipelineDefinition)
    assert foo_pipeline.name == 'foo'

    assert 'foo' in called
    assert called['foo'] == 1

    pipelines = repo.get_all_pipelines()

    assert set(['foo', 'bar']) == set([pipeline.name for pipeline in pipelines])



import itertools
import json
import uuid

from collections import (
    OrderedDict,
    namedtuple,
)
from contextlib import contextmanager

from dagster import check
from dagster.utils.logging import (
    CompositeLogger,
    INFO,
    define_colored_console_logger,
)

Metric = namedtuple('Metric', 'context_dict metric_name value')


def _kv_message(all_items):
    return ' '.join(
        ['{key}={value}'.format(key=key, value=json.dumps(value)) for key, value in all_items]
    )


class ExecutionContext(object):
    '''
    A context object flowed through the entire scope of single execution of a
    pipeline of solids. This is used by both framework and user code to log
    messages and metrics. It also maintains a stack of context values so that
    logs, metrics, and any future reporting are reported with a minimal, consistent
    level of context so that developers do not have to repeatedly log well-known
    information (e.g. the name of the solid, the name of the pipeline, etc) when
    logging. Additionally tool author may add their own context values to assist
    reporting.


    Args:
        loggers (List[logging.Logger]):
            The list of loggers that will be invoked whenever the context logging
            functions are called.

        resources(Any):
            An arbitrary user-defined object that can be passed in by a user and
            then access during pipeline execution. This exists so that a user can
            inject their own objects into the context without having to subclass
            ExecutionContext.
    '''

    def __init__(self, loggers=None, resources=None):
        self._logger = CompositeLogger(loggers=loggers)
        self._context_dict = OrderedDict()
        self._metrics = []
        self.resources = resources

    @staticmethod
    def console_logging(log_level=INFO, resources=None):
        return ExecutionContext(
            loggers=[define_colored_console_logger('dagster', log_level)],
            resources=resources,
        )

    def _log(self, method, orig_message, message_props):
        check.str_param(method, 'method')
        check.str_param(orig_message, 'orig_message')
        check.dict_param(message_props, 'message_props')

        check.invariant(
            'extra' not in message_props,
            'do not allow until explicit support is handled',
        )
        check.invariant(
            'exc_info' not in message_props,
            'do not allow until explicit support is handled',
        )

        check.invariant('orig_message' not in message_props, 'orig_message reserved value')
        check.invariant('message' not in message_props, 'message reserved value')
        check.invariant('log_message_id' not in message_props, 'log_message_id reserved value')

        log_message_id = str(uuid.uuid4())

        synth_props = {'orig_message': orig_message, 'log_message_id': log_message_id}

        # We first generate all props for the purpose of producing the semi-structured
        # log message via _kv_messsage
        all_props = dict(
            itertools.chain(
                synth_props.items(),
                self._context_dict.items(),
                message_props.items(),
            )
        )

        message_with_structured_props = _kv_message(all_props.items())

        getattr(self._logger, method)(message_with_structured_props, extra=all_props)

    def debug(self, msg, **kwargs):
        '''
        Debug level logging directive. Ends up invoking loggers with DEBUG error level.

        The message will be automatically adorned with context information about the name
        of the pipeline, the name of the solid, and so forth. The use can also add
        context values during execution using the value() method of ExecutionContext.
        Therefore it is generally unnecessary to include this type of information
        (solid name, pipeline name, etc) in the log message unless it is critical
        for the readability/fluency of the log message text itself.

        You can optionally additional context key-value pairs to an individual log
        message using the keyword args to this message

        Args:
            msg (str): The core string
            **kwargs (Dict[str, Any]): Additional context values for only this log message.
        '''
        return self._log('debug', msg, kwargs)

    def info(self, msg, **kwargs):
        '''Log at INFO level

        See debug()'''
        return self._log('info', msg, kwargs)

    def warning(self, msg, **kwargs):
        '''Log at WARNING level

        See debug()'''
        return self._log('warning', msg, kwargs)

    def error(self, msg, **kwargs):
        '''Log at ERROR level

        See debug()'''
        return self._log('error', msg, kwargs)

    def critical(self, msg, **kwargs):
        '''Log at CRITICAL level

        See debug()'''
        return self._log('critical', msg, kwargs)

    # FIXME: Actually make this work
    # def exception(self, e):
    #     check.inst_param(e, 'e', Exception)

    #     # this is pretty lame right. should embellish with more data (stack trace?)
    #     return self._log('error', str(e))

    @contextmanager
    def value(self, key, value):
        '''
        Adds a context value to the Execution for a particular scope, using the
        python contextmanager abstraction. This allows the user to add scoped metadata
        just like the framework does (for things such as solid name).

        Examples:

        .. code-block:: python

            with context.value('some_key', 'some_value):
                context.info('msg with some_key context value')

            context.info('msg without some_key context value')

        '''

        check.str_param(key, 'key')
        check.not_none_param(value, 'value')
        with self.values({key: value}):
            yield

    @contextmanager
    def values(self, ddict):
        check.dict_param(ddict, 'ddict')

        for key, value in ddict.items():
            check.invariant(not key in self._context_dict, 'Should not be in context')
            self._context_dict[key] = value

        yield

        for key in ddict.keys():
            self._context_dict.pop(key)


class TransformExecutionInfo(namedtuple('_TransformExecutionInfo', 'context config')):
    '''An instance of TransformExecutionInfo is passed every solid transform function.

    Attributes:

        context (ExecutionContext): Context instance for this pipeline invocation
        config (Any): Config object for current solid
    '''

    def __new__(cls, context, config):
        return super(TransformExecutionInfo, cls).__new__(
            cls,
            check.inst_param(context, 'context', ExecutionContext),
            config,
        )



import pytest

from dagster import (
    ConfigDefinition,
    ExecutionContext,
    Field,
    OutputDefinition,
    PipelineDefinition,
    PipelineContextDefinition,
    config,
    execute_pipeline,
    lambda_solid,
    solid,
    types,
)
from dagster.core.errors import (DagsterTypeError, DagsterInvariantViolationError)
from dagster.utils.logging import (INFO, ERROR)

# protected variable. need to test loggers
# pylint: disable=W0212


def test_default_context():
    called = {}

    @solid(inputs=[], outputs=[OutputDefinition()])
    def default_context_transform(info):
        called['yes'] = True
        for logger in info.context._logger.loggers:
            assert logger.level == INFO

    pipeline = PipelineDefinition(solids=[default_context_transform])
    execute_pipeline(pipeline)

    assert called['yes']


def test_default_context_with_log_level():
    @solid(inputs=[], outputs=[OutputDefinition()])
    def default_context_transform(info):
        for logger in info.context._logger.loggers:
            assert logger.level == INFO

    pipeline = PipelineDefinition(solids=[default_context_transform])
    execute_pipeline(
        pipeline,
        environment=config.Environment(context=config.Context(config={'log_level': 'INFO'}))
    )

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            pipeline,
            environment=config.Environment(context=config.Context(config={'log_level': 2}))
        )


def test_default_value():
    def _get_config_test_solid(config_key, config_value):
        @solid(inputs=[], outputs=[OutputDefinition()])
        def config_test(info):
            assert info.context.resources == {config_key: config_value}

        return config_test

    pipeline = PipelineDefinition(
        solids=[_get_config_test_solid('field_one', 'heyo')],
        context_definitions={
            'custom_one':
            PipelineContextDefinition(
                config_def=ConfigDefinition.config_dict(
                    {
                        'field_one':
                        Field(
                            dagster_type=types.String,
                            is_optional=True,
                            default_value='heyo',
                        )
                    }
                ),
                context_fn=lambda info: ExecutionContext(resources=info.config),
            ),
        }
    )

    execute_pipeline(
        pipeline, environment=config.Environment(context=config.Context('custom_one', {}))
    )


def test_custom_contexts():
    @solid(inputs=[], outputs=[OutputDefinition()])
    def custom_context_transform(info):
        assert info.context.resources == {'field_one': 'value_two'}

    pipeline = PipelineDefinition(
        solids=[custom_context_transform],
        context_definitions={
            'custom_one':
            PipelineContextDefinition(
                config_def=ConfigDefinition.config_dict(
                    {
                        'field_one': Field(dagster_type=types.String)
                    }
                ),
                context_fn=lambda info: ExecutionContext(resources=info.config),
            ),
            'custom_two':
            PipelineContextDefinition(
                config_def=ConfigDefinition.config_dict(
                    {
                        'field_one': Field(dagster_type=types.String)
                    }
                ),
                context_fn=lambda info: ExecutionContext(resources=info.config),
            )
        },
    )

    environment_one = config.Environment(
        context=config.Context('custom_one', {'field_one': 'value_two'})
    )

    execute_pipeline(pipeline, environment=environment_one)

    environment_two = config.Environment(
        context=config.Context('custom_two', {'field_one': 'value_two'})
    )

    execute_pipeline(pipeline, environment=environment_two)


def test_yield_context():
    events = []

    @solid(inputs=[], outputs=[OutputDefinition()])
    def custom_context_transform(info):
        assert info.context.resources == {'field_one': 'value_two'}
        assert info.context._context_dict['foo'] == 'bar'  # pylint: disable=W0212
        events.append('during')

    def _yield_context(info):
        events.append('before')
        context = ExecutionContext(resources=info.config)
        with context.value('foo', 'bar'):
            yield context
        events.append('after')

    pipeline = PipelineDefinition(
        solids=[custom_context_transform],
        context_definitions={
            'custom_one':
            PipelineContextDefinition(
                config_def=ConfigDefinition.config_dict(
                    {
                        'field_one': Field(dagster_type=types.String)
                    }
                ),
                context_fn=_yield_context,
            ),
        }
    )

    environment_one = config.Environment(
        context=config.Context('custom_one', {'field_one': 'value_two'})
    )

    execute_pipeline(pipeline, environment=environment_one)

    assert events == ['before', 'during', 'after']


# TODO: reenable pending the ability to specific optional arguments
# https://github.com/dagster-io/dagster/issues/56
def test_invalid_context():
    @lambda_solid
    def never_transform():
        raise Exception('should never execute')

    default_context_pipeline = PipelineDefinition(solids=[never_transform])

    environment_context_not_found = config.Environment(context=config.Context('not_found', {}))

    with pytest.raises(DagsterInvariantViolationError, message='Context not_found not found'):
        execute_pipeline(
            default_context_pipeline,
            environment=environment_context_not_found,
            throw_on_error=True
        )

    environment_field_name_mismatch = config.Environment(
        context=config.Context(config={'unexpected': 'value'})
    )

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            default_context_pipeline,
            environment=environment_field_name_mismatch,
            throw_on_error=True
        )

    with_argful_context_pipeline = PipelineDefinition(
        solids=[never_transform],
        context_definitions={
            'default':
            PipelineContextDefinition(
                config_def=ConfigDefinition.config_dict({
                    'string_field': Field(types.String)
                }),
                context_fn=lambda info: info.config,
            )
        }
    )

    environment_no_config_error = config.Environment(context=config.Context(config={}))

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            with_argful_context_pipeline,
            environment=environment_no_config_error,
            throw_on_error=True
        )

    environment_type_mismatch_error = config.Environment(
        context=config.Context(config={'string_field': 1})
    )

    with pytest.raises(DagsterTypeError, message='Argument mismatch in context default'):
        execute_pipeline(
            with_argful_context_pipeline,
            environment=environment_type_mismatch_error,
            throw_on_error=True
        )



from __future__ import print_function
import logging
import re
import textwrap

import click

from dagster import (
    RepositoryDefinition,
    PipelineDefinition,
    check,
    config,
)

from dagster.core.definitions import ExecutionGraph
from dagster.core.execution import execute_pipeline_iterator
from dagster.graphviz import build_graphviz_graph
from dagster.utils import load_yaml_from_path
from dagster.utils.indenting_printer import IndentingPrinter

from .dynamic_loader import (
    PipelineTargetInfo,
    RepositoryTargetInfo,
    load_pipeline_from_target_info,
    load_repository_from_file,
    load_repository_from_target_info,
    pipeline_target_command,
    repository_config_argument,
    load_target_info_from_kwargs,
)


def create_pipeline_cli():
    group = click.Group(name="pipeline")
    group.add_command(list_command)
    group.add_command(print_command)
    group.add_command(graphviz_command)
    group.add_command(execute_command)
    return group


REPO_TARGET_WARNING = (
    'Can only use ONE of --repository-yaml/-y, --python-file/-f, --module-name/-m.'
)


@click.command(
    name='list',
    help="List the pipelines in a repository. {warning}".format(warning=REPO_TARGET_WARNING),
)
@repository_config_argument
def list_command(**kwargs):
    repository_target_info = load_target_info_from_kwargs(kwargs)
    repository = load_repository_from_target_info(repository_target_info)

    title = 'Repository {name}'.format(name=repository.name)
    click.echo(title)
    click.echo('*' * len(title))
    first = True
    for pipeline in repository.get_all_pipelines():
        pipeline_title = 'Pipeline: {name}'.format(name=pipeline.name)

        if not first:
            click.echo('*' * len(pipeline_title))
        first = False

        click.echo(pipeline_title)
        if pipeline.description:
            click.echo('Description:')
            click.echo(format_description(pipeline.description, indent=' ' * 4))
        click.echo('Solids: (Execution Order)')
        solid_graph = ExecutionGraph(pipeline, pipeline.solids, pipeline.dependency_structure)
        for solid in solid_graph.topological_solids:
            click.echo('    ' + solid.name)


def format_description(desc, indent):
    check.str_param(desc, 'desc')
    check.str_param(indent, 'indent')
    desc = re.sub(r'\s+', ' ', desc)
    dedented = textwrap.dedent(desc)
    wrapper = textwrap.TextWrapper(initial_indent='', subsequent_indent=indent)
    filled = wrapper.fill(dedented)
    return filled


def create_pipeline_from_cli_args(kwargs):
    check.dict_param(kwargs, 'kwargs')

    pipeline_names = list(kwargs['pipeline_name'])

    if not pipeline_names:
        pipeline_name = None
    elif len(pipeline_names) == 1:
        pipeline_name = pipeline_names[0]
    else:
        check.failed('Can only handle zero or one pipeline args')

    if kwargs['pipeline_name'] and kwargs['repository_yaml'] is None:
        repository_yaml = 'repository.yml'
    else:
        repository_yaml = kwargs['repository_yaml']

    return load_pipeline_from_target_info(
        PipelineTargetInfo(
            repository_yaml=repository_yaml,
            pipeline_name=pipeline_name,
            python_file=kwargs['python_file'],
            pipeline_fn_name=kwargs['pipeline_fn_name'],
            module_name=kwargs['module_name'],
            repository_fn_name=kwargs['repository_fn_name'],
        )
    )


def get_pipeline_instructions(command_name):
    return (
        'This commands targets a pipeline. The pipeline can be specified in a number of ways:'
        '\n\n1. dagster {command_name} <<pipeline_name>> (works if .repository.yml exists)'
        '\n\n2. dagster {command_name} <<pipeline_name>> -y path/to/repository.yml'
        '\n\n3. dagster {command_name} -f /path/to/file.py -p define_some_pipeline'
        '\n\n4. dagster {command_name} -m a_module.submodule  -p define_some_pipeline'
        '\n\n5. dagster {command_name} -f /path/to/file.py -r define_some_repo -p pipeline_name'
        '\n\n6. dagster {command_name} -m a_module.submodule  -r define_some_repo -p pipeline_name'
    ).format(command_name=command_name)


@click.command(
    name='print',
    help='Print a pipeline.\n\n{instructions}'.format(
        instructions=get_pipeline_instructions('print')
    ),
)
@click.option('--verbose', is_flag=True)
@pipeline_target_command
def print_command(verbose, **kwargs):
    pipeline = create_pipeline_from_cli_args(kwargs)

    if verbose:
        print_pipeline(pipeline, full=True, print_fn=click.echo)
    else:
        print_solids(pipeline, print_fn=click.echo)


def print_solids(pipeline, print_fn):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.callable_param(print_fn, 'print_fn')

    printer = IndentingPrinter(indent_level=2, printer=print_fn)
    printer.line('Pipeline: {name}'.format(name=pipeline.name))

    printer.line('Solids:')
    for solid in pipeline.solids:
        with printer.with_indent():
            printer.line('Solid: {name}'.format(name=solid.name))


def print_pipeline(pipeline, full, print_fn):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.bool_param(full, 'full')
    check.callable_param(print_fn, 'print_fn')

    printer = IndentingPrinter(indent_level=2, printer=print_fn)
    printer.line('Pipeline: {name}'.format(name=pipeline.name))
    print_description(printer, pipeline.description)

    if not full:
        return

    with printer.with_indent():
        printer.line('Context Definitions:')

        with printer.with_indent():

            for context_name, context_definition in pipeline.context_definitions.items():
                print_context_definition(printer, context_name, context_definition)

    printer.line('Solids:')
    for solid in pipeline.solids:
        with printer.with_indent():
            print_solid(printer, solid)


def print_description(printer, desc):
    with printer.with_indent():
        if desc:
            printer.line('Description:')
            with printer.with_indent():
                printer.line(format_description(desc, printer.current_indent_str))


def print_context_definition(printer, context_name, context_definition):
    printer.line('Name: {context_name}'.format(context_name=context_name))

    print_description(printer, context_definition.description)

    printer.line(
        'Type: {config_type}'.format(config_type=context_definition.config_def.config_type.name)
    )


def print_solid(printer, solid):
    printer.line('Solid: {name}'.format(name=solid.name))

    with printer.with_indent():
        print_inputs(printer, solid)

        printer.line('Outputs:')

        for output_def in solid.output_defs:
            print(output_def.name)


def print_inputs(printer, solid):
    printer.line('Inputs:')
    for input_def in solid.input_defs:
        with printer.with_indent():
            printer.line('Input: {name}'.format(name=input_def.name))


def format_argument_dict(arg_def_dict):
    return ', '.join(
        [
            '{name}: {type}'.format(name=name, type=arg_def.dagster_type.name)
            for name, arg_def in arg_def_dict.items()
        ]
    )


@click.command(
    name='graphviz',
    help=(
        'Visualize a pipeline using graphviz. Must be installed on your system '
        '(e.g. homebrew install graphviz on mac). \n\n{instructions}'.format(
            instructions=get_pipeline_instructions('graphviz')
        )
    ),
)
@click.option('--only-solids', is_flag=True)
@pipeline_target_command
def graphviz_command(only_solids, **kwargs):
    pipeline_target_info = PipelineTargetInfo(**kwargs)
    pipeline = load_pipeline_from_target_info(pipeline_target_info)
    build_graphviz_graph(pipeline, only_solids).view(cleanup=True)


LOGGING_DICT = {
    'DEBUG': logging.DEBUG,
    'INFO': logging.INFO,
    'WARN': logging.WARN,
    'ERROR': logging.ERROR,
    'CRITICAL': logging.CRITICAL,
}


@click.command(
    name='execute',
    help='Execute a pipeline.\n\n{instructions}'.format(
        instructions=get_pipeline_instructions('execute')
    ),
)
@pipeline_target_command
@click.option(
    '-e',
    '--env',
    type=click.Path(
        exists=True,
        file_okay=True,
        dir_okay=False,
        readable=True,
        resolve_path=True,
    ),
)
def execute_command(env, **kwargs):
    pipeline_target_info = PipelineTargetInfo(**kwargs)
    pipeline = load_pipeline_from_target_info(pipeline_target_info)
    do_execute_command(pipeline, env, print)


def do_execute_command(pipeline, env, printer):
    check.inst_param(pipeline, 'pipeline', PipelineDefinition)
    check.str_param(env, 'env')
    check.callable_param(printer, 'printer')

    env_config = load_yaml_from_path(env)

    environment = config.construct_environment(env_config)

    pipeline_iter = execute_pipeline_iterator(pipeline, environment)

    process_results_for_console(pipeline_iter)


def process_results_for_console(pipeline_iter):
    results = []

    for result in pipeline_iter:
        if not result.success:
            result.reraise_user_error()
        results.append(result)

    return results



from six import (string_types, integer_types)

from dagster import check

from dagster.core.errors import DagsterEvaluateValueError


class DagsterType(object):
    '''Base class for Dagster Type system. Should be inherited by a subclass.
    Subclass must implement `evaluate_value`

    Attributes:
      name (str): Name of the type

      description (str): Description of the type
    '''

    def __init__(self, name, description=None):
        self.name = check.str_param(name, 'name')
        self.description = check.opt_str_param(description, 'description')
        self.__doc__ = description

    def __repr__(self):
        return 'DagsterType({name})'.format(name=self.name)

    def evaluate_value(self, _value):
        '''Subclasses can implement this method. Check if the value is a valid one
        and return a processed version of it. If value is invalid,
        raise `DagsterEvaluateValueError`.

        This class provides a default implementation of this method

        Args:
          value: The value to check

        Returns:
          value: A transformed value
        '''
        check.not_implemented('Must implement in subclass')


class UncoercedTypeMixin(object):
    '''This is a helper mixin used when you only want to do a type check
    against an in-memory value and then leave that value uncoerced. Only
    is_python_valid_value must be implemented for these classes.
    evaluate_value is implemented for you.
    '''

    def is_python_valid_value(self, _value):
        '''Subclasses must implement this method. Check if the value and output a boolean.

        Returns:
          bool: Whether the value is valid.
        '''
        check.failed('must implement')

    def evaluate_value(self, value):
        if not self.is_python_valid_value(value):
            raise DagsterEvaluateValueError(
                'Expected valid value for {type_name} but got {value}'.format(
                    type_name=self.name, value=repr(value)
                )
            )
        return value


class DagsterScalarType(UncoercedTypeMixin, DagsterType):
    '''Base class for dagster types that are scalar python values.

    Attributes:
      name (str): Name of the type

      description (str): Description of the type
    '''

    def __init__(self, *args, **kwargs):
        super(DagsterScalarType, self).__init__(*args, **kwargs)


class _DagsterAnyType(UncoercedTypeMixin, DagsterType):
    def __init__(self):
        super(_DagsterAnyType, self).__init__(
            name='Any', description='The type that allows any value, including no value.'
        )

    def is_python_valid_value(self, _value):
        return True


def nullable_isinstance(value, typez):
    return value is None or isinstance(value, typez)


class PythonObjectType(UncoercedTypeMixin, DagsterType):
    '''Dagster Type that checks if the value is an instance of some `python_type`'''

    def __init__(
        self,
        name,
        python_type,
        description=None,
    ):
        super(PythonObjectType, self).__init__(name, description)
        self.python_type = check.type_param(python_type, 'python_type')

    def is_python_valid_value(self, value):
        return nullable_isinstance(value, self.python_type)


class _DagsterStringType(DagsterScalarType):
    def is_python_valid_value(self, value):
        return nullable_isinstance(value, string_types)


class _DagsterIntType(DagsterScalarType):
    def __init__(self):
        super(_DagsterIntType, self).__init__('Int', description='An integer.')

    def is_python_valid_value(self, value):
        if isinstance(value, bool):
            return False
        return nullable_isinstance(value, integer_types)


class _DagsterBoolType(DagsterScalarType):
    def __init__(self):
        super(_DagsterBoolType, self).__init__('Bool', description='A boolean.')

    def is_python_valid_value(self, value):
        return nullable_isinstance(value, bool)


class __FieldValueSentinel:
    pass


FIELD_NO_DEFAULT_PROVIDED = __FieldValueSentinel


class Field:
    '''
    A Field in a DagsterCompositeType.

    Attributes:
        dagster_type (DagsterType): The type of the field.
        default_value (Any):
            If the Field is optional, a default value can be provided when the field value
            is not specified.
        is_optional (bool): Is the field optional.
        description (str): Description of the field.
    '''

    def __init__(
        self,
        dagster_type,
        default_value=FIELD_NO_DEFAULT_PROVIDED,
        is_optional=False,
        description=None
    ):
        if not is_optional:
            check.param_invariant(
                default_value == FIELD_NO_DEFAULT_PROVIDED,
                'default_value',
                'required arguments should not specify default values',
            )

        self.dagster_type = check.inst_param(dagster_type, 'dagster_type', DagsterType)
        self.description = check.opt_str_param(description, 'description')
        self.is_optional = check.bool_param(is_optional, 'is_optional')
        self.default_value = default_value

    @property
    def default_provided(self):
        '''Was a default value provided

        Returns:
            bool: Yes or no
        '''
        return self.default_value != FIELD_NO_DEFAULT_PROVIDED


class FieldDefinitionDictionary(dict):
    def __init__(self, ddict):
        check.dict_param(ddict, 'ddict', key_type=str, value_type=Field)
        super(FieldDefinitionDictionary, self).__init__(ddict)

    def __setitem__(self, _key, _value):
        check.failed('This dictionary is readonly')


class DagsterCompositeType(DagsterType):
    '''Dagster type representing a type with a list of named :py:class:`Field` objects.
    '''

    def __init__(self, name, fields, ctor, description=None):
        self.field_dict = FieldDefinitionDictionary(fields)
        self.ctor = check.callable_param(ctor, 'ctor')
        super(DagsterCompositeType, self).__init__(name, description)

    def evaluate_value(self, value):
        if value is not None and not isinstance(value, dict):
            raise DagsterEvaluateValueError('Incoming value for composite must be dict')
        return process_incoming_composite_value(self, value, self.ctor)


class ConfigDictionary(DagsterCompositeType):
    '''Configuration dictionary.

    Typed-checked but then passed to implementations as a python dict

    Arguments:
      fields (dict): dictonary of :py:class:`Field` objects keyed by name'''

    def __init__(self, fields):
        super(ConfigDictionary, self).__init__(
            'ConfigDictionary',
            fields,
            lambda val: val,
            self.__doc__,
        )


def process_incoming_composite_value(dagster_composite_type, incoming_value, ctor):
    check.inst_param(dagster_composite_type, 'dagster_composite_type', DagsterCompositeType)
    incoming_value = check.opt_dict_param(incoming_value, 'incoming_value', key_type=str)
    check.callable_param(ctor, 'ctor')

    field_dict = dagster_composite_type.field_dict

    defined_args = set(field_dict.keys())
    received_args = set(incoming_value.keys())

    for received_arg in received_args:
        if received_arg not in defined_args:
            raise DagsterEvaluateValueError(
                'Field {received} not found. Defined fields: {defined}'.format(
                    defined=repr(defined_args),
                    received=received_arg,
                )
            )

    for expected_field, field_def in field_dict.items():
        if field_def.is_optional:
            continue

        check.invariant(not field_def.default_provided)

        if expected_field not in received_args:
            raise DagsterEvaluateValueError(
                'Did not not find {expected}. Defined fields: {defined}'.format(
                    expected=expected_field,
                    defined=repr(defined_args),
                )
            )

    fields_to_pass = {}

    for expected_field, field_def in field_dict.items():
        if expected_field in received_args:
            evaluation_result = field_def.dagster_type.evaluate_value(
                incoming_value[expected_field]
            )
            fields_to_pass[expected_field] = evaluation_result
        elif field_def.default_provided:
            fields_to_pass[expected_field] = field_def.default_value
        else:
            check.invariant(field_def.is_optional and not field_def.default_provided)

    return ctor(fields_to_pass)


String = _DagsterStringType(name='String', description='A string.')
Path = _DagsterStringType(
    name='Path',
    description='''
A string the represents a path. It is very useful for some tooling
to know that a string indeed represents a file path. That way they
can, for example, make the paths relative to a different location
for a particular execution environment.
''',
)
Int = _DagsterIntType()
Bool = _DagsterBoolType()
Any = _DagsterAnyType()
Dict = PythonObjectType('Dict', dict)



import pytest

from dagster import (RepositoryDefinition, PipelineDefinition)

from dagster.cli.dynamic_loader import (
    FileTargetFunction,
    InvalidPipelineLoadingComboError,
    PipelineLoadingModeData,
    ModuleTargetFunction,
    PipelineTargetInfo,
    PipelineTargetMode,
    RepositoryModuleData,
    RepositoryPythonFileData,
    RepositoryLoadingModeData,
    RepositoryTargetInfo,
    RepositoryTargetMode,
    RepositoryYamlData,
    create_repository_loading_mode_data,
    create_pipeline_loading_mode_data,
    load_pipeline_from_target_info,
    load_repository_object_from_target_info,
)

from dagster.utils import script_relative_path


def test_repository_python_file():
    mode_data = create_pipeline_loading_mode_data(
        PipelineTargetInfo(
            module_name=None,
            pipeline_fn_name=None,
            pipeline_name='foo',
            python_file='bar_repo.py',
            repository_fn_name='define_bar_repo',
            repository_yaml=None,
        )
    )

    assert mode_data == PipelineLoadingModeData(
        mode=PipelineTargetMode.REPOSITORY_PYTHON_FILE,
        data=RepositoryPythonFileData(
            file_target_function=FileTargetFunction(
                python_file='bar_repo.py',
                fn_name='define_bar_repo',
            ),
            pipeline_name='foo',
        )
    )

    with pytest.raises(InvalidPipelineLoadingComboError):
        create_pipeline_loading_mode_data(
            PipelineTargetInfo(
                module_name='kdjfkd',
                pipeline_fn_name=None,
                pipeline_name='foo',
                python_file='bar_repo.py',
                repository_fn_name='define_bar_repo',
                repository_yaml=None,
            )
        )

    with pytest.raises(InvalidPipelineLoadingComboError):
        create_pipeline_loading_mode_data(
            PipelineTargetInfo(
                module_name=None,
                pipeline_fn_name='kjdfd',
                pipeline_name='foo',
                python_file='bar_repo.py',
                repository_fn_name='define_bar_repo',
                repository_yaml=None,
            )
        )

    with pytest.raises(InvalidPipelineLoadingComboError):
        create_pipeline_loading_mode_data(
            PipelineTargetInfo(
                module_name=None,
                pipeline_fn_name=None,
                pipeline_name='foo',
                python_file='bar_repo.py',
                repository_fn_name='define_bar_repo',
                repository_yaml='kjdfkdjf',
            )
        )


def test_repository_module():
    mode_data = create_pipeline_loading_mode_data(
        PipelineTargetInfo(
            module_name='bar_module',
            pipeline_fn_name=None,
            pipeline_name='foo',
            python_file=None,
            repository_fn_name='define_bar_repo',
            repository_yaml=None,
        )
    )

    assert mode_data == PipelineLoadingModeData(
        mode=PipelineTargetMode.REPOSITORY_MODULE,
        data=RepositoryModuleData(
            module_target_function=ModuleTargetFunction(
                module_name='bar_module',
                fn_name='define_bar_repo',
            ),
            pipeline_name='foo',
        )
    )


def test_pipeline_python_file():
    mode_data = create_pipeline_loading_mode_data(
        PipelineTargetInfo(
            module_name=None,
            pipeline_fn_name='define_pipeline',
            pipeline_name=None,
            python_file='foo.py',
            repository_fn_name=None,
            repository_yaml=None,
        )
    )

    assert mode_data == PipelineLoadingModeData(
        mode=PipelineTargetMode.PIPELINE_PYTHON_FILE,
        data=FileTargetFunction(
            python_file='foo.py',
            fn_name='define_pipeline',
        ),
    )


def test_pipeline_module():
    mode_data = create_pipeline_loading_mode_data(
        PipelineTargetInfo(
            module_name='foo_module',
            pipeline_fn_name='define_pipeline',
            pipeline_name=None,
            python_file=None,
            repository_fn_name=None,
            repository_yaml=None,
        )
    )

    assert mode_data == PipelineLoadingModeData(
        mode=PipelineTargetMode.PIPELINE_MODULE,
        data=ModuleTargetFunction(
            module_name='foo_module',
            fn_name='define_pipeline',
        ),
    )


def test_yaml_file():
    assert create_pipeline_loading_mode_data(
        PipelineTargetInfo(
            module_name=None,
            pipeline_fn_name=None,
            pipeline_name='foobar',
            python_file=None,
            repository_fn_name=None,
            repository_yaml='some_file.yml',
        )
    ) == PipelineLoadingModeData(
        mode=PipelineTargetMode.REPOSITORY_YAML_FILE,
        data=RepositoryYamlData(repository_yaml='some_file.yml', pipeline_name='foobar')
    )

    with pytest.raises(InvalidPipelineLoadingComboError):
        assert create_pipeline_loading_mode_data(
            PipelineTargetInfo(
                module_name='kdjfdk',
                pipeline_fn_name=None,
                pipeline_name='foobar',
                python_file=None,
                repository_fn_name=None,
                repository_yaml=None,
            )
        )

    with pytest.raises(InvalidPipelineLoadingComboError):
        assert create_pipeline_loading_mode_data(
            PipelineTargetInfo(
                module_name=None,
                pipeline_fn_name='kjdfkd',
                pipeline_name='foobar',
                python_file=None,
                repository_fn_name=None,
                repository_yaml=None,
            )
        )

    with pytest.raises(InvalidPipelineLoadingComboError):
        assert create_pipeline_loading_mode_data(
            PipelineTargetInfo(
                module_name=None,
                pipeline_fn_name=None,
                pipeline_name='foobar',
                python_file='kjdfkdj',
                repository_fn_name=None,
                repository_yaml=None,
            )
        )

    with pytest.raises(InvalidPipelineLoadingComboError):
        assert create_pipeline_loading_mode_data(
            PipelineTargetInfo(
                module_name=None,
                pipeline_fn_name=None,
                pipeline_name='foobar',
                python_file=None,
                repository_fn_name='kjdkfjd',
                repository_yaml=None,
            )
        )


def test_load_from_repository_file():
    pipeline = load_pipeline_from_target_info(
        PipelineTargetInfo(
            module_name=None,
            pipeline_fn_name=None,
            pipeline_name='foo',
            python_file=script_relative_path('test_dynamic_loader.py'),
            repository_fn_name='define_bar_repo',
            repository_yaml=None,
        )
    )

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'


def test_load_from_repository_module():
    pipeline = load_pipeline_from_target_info(
        PipelineTargetInfo(
            module_name='dagster.cli.cli_tests.test_dynamic_loader',
            pipeline_fn_name=None,
            pipeline_name='foo',
            python_file=None,
            repository_fn_name='define_bar_repo',
            repository_yaml=None,
        )
    )

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'


def test_load_from_pipeline_file():
    pipeline = load_pipeline_from_target_info(
        PipelineTargetInfo(
            module_name=None,
            pipeline_fn_name='define_foo_pipeline',
            pipeline_name=None,
            python_file=script_relative_path('test_dynamic_loader.py'),
            repository_fn_name=None,
            repository_yaml=None,
        )
    )

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'


def test_load_from_pipeline_module():
    pipeline = load_pipeline_from_target_info(
        PipelineTargetInfo(
            module_name='dagster.cli.cli_tests.test_dynamic_loader',
            pipeline_fn_name='define_foo_pipeline',
            pipeline_name=None,
            python_file=None,
            repository_fn_name=None,
            repository_yaml=None,
        )
    )

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'


def test_loader_from_default_repository_module_yaml():
    pipeline = load_pipeline_from_target_info(
        PipelineTargetInfo(
            module_name=None,
            pipeline_fn_name=None,
            pipeline_name='foo',
            python_file=None,
            repository_fn_name=None,
            repository_yaml=script_relative_path('repository_module.yml'),
        )
    )

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'


def test_loader_from_default_repository_file_yaml():
    pipeline = load_pipeline_from_target_info(
        PipelineTargetInfo(
            module_name=None,
            pipeline_fn_name=None,
            pipeline_name='foo',
            python_file=None,
            repository_fn_name=None,
            repository_yaml=script_relative_path('repository_file.yml'),
        )
    )

    assert isinstance(pipeline, PipelineDefinition)
    assert pipeline.name == 'foo'


def test_repo_mode_data():
    assert create_repository_loading_mode_data(
        RepositoryTargetInfo(
            repository_yaml='foo.yaml',
            module_name=None,
            python_file=None,
            fn_name=None,
        )
    ) == RepositoryLoadingModeData(
        mode=RepositoryTargetMode.YAML_FILE,
        data='foo.yaml',
    )

    assert create_repository_loading_mode_data(
        RepositoryTargetInfo(
            repository_yaml=None,
            module_name='foo',
            python_file=None,
            fn_name='define_bar_repo',
        )
    ) == RepositoryLoadingModeData(
        mode=RepositoryTargetMode.MODULE,
        data=ModuleTargetFunction(
            module_name='foo',
            fn_name='define_bar_repo',
        )
    )

    assert create_repository_loading_mode_data(
        RepositoryTargetInfo(
            repository_yaml=None,
            module_name=None,
            python_file='foo.py',
            fn_name='define_bar_repo',
        )
    ) == RepositoryLoadingModeData(
        mode=RepositoryTargetMode.FILE,
        data=FileTargetFunction(
            python_file='foo.py',
            fn_name='define_bar_repo',
        )
    )


def test_repo_yaml_module_dynamic_load():
    repository = load_repository_object_from_target_info(
        RepositoryTargetInfo(
            repository_yaml=script_relative_path('repository_module.yml'),
            module_name=None,
            python_file=None,
            fn_name=None,
        )
    ).fn()

    assert isinstance(repository, RepositoryDefinition)
    assert repository.name == 'bar'


def test_repo_yaml_file_dynamic_load():
    repository = load_repository_object_from_target_info(
        RepositoryTargetInfo(
            repository_yaml=script_relative_path('repository_file.yml'),
            module_name=None,
            python_file=None,
            fn_name=None,
        )
    ).fn()

    assert isinstance(repository, RepositoryDefinition)
    assert repository.name == 'bar'


def test_repo_module_dynamic_load():
    repository = load_repository_object_from_target_info(
        RepositoryTargetInfo(
            repository_yaml=None,
            module_name='dagster.cli.cli_tests.test_dynamic_loader',
            python_file=None,
            fn_name='define_bar_repo',
        )
    ).fn()

    assert isinstance(repository, RepositoryDefinition)
    assert repository.name == 'bar'


def test_repo_file_dynamic_load():
    repository = load_repository_object_from_target_info(
        RepositoryTargetInfo(
            repository_yaml=None,
            module_name=None,
            python_file=script_relative_path('test_dynamic_loader.py'),
            fn_name='define_bar_repo',
        )
    ).fn()

    assert isinstance(repository, RepositoryDefinition)
    assert repository.name == 'bar'


def test_repo_module_dynamic_load_from_pipeline():
    repository = load_repository_object_from_target_info(
        RepositoryTargetInfo(
            repository_yaml=None,
            module_name='dagster.cli.cli_tests.test_dynamic_loader',
            python_file=None,
            fn_name='define_foo_pipeline',
        )
    ).fn()

    assert isinstance(repository, RepositoryDefinition)
    assert repository.name == '<<unnamed>>'
    assert repository.get_pipeline('foo').name == 'foo'


def test_repo_file_dynamic_load_from_pipeline():
    repository = load_repository_object_from_target_info(
        RepositoryTargetInfo(
            repository_yaml=None,
            module_name=None,
            python_file=script_relative_path('test_dynamic_loader.py'),
            fn_name='define_foo_pipeline',
        )
    ).fn()

    assert isinstance(repository, RepositoryDefinition)
    assert repository.name == '<<unnamed>>'
    assert repository.get_pipeline('foo').name == 'foo'


def define_foo_pipeline():
    return PipelineDefinition(name='foo', solids=[])


def define_bar_repo():
    return RepositoryDefinition('bar', {'foo': define_foo_pipeline})



import pytest

from dagster import (
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    PipelineDefinition,
    SolidDefinition,
    OutputDefinition,
    InputDefinition,
)


def solid_a_b_list():
    return [
        SolidDefinition(
            name='A',
            inputs=[],
            outputs=[OutputDefinition()],
            transform_fn=lambda _info, _inputs: None,
        ),
        SolidDefinition(
            name='B',
            inputs=[InputDefinition('b_input')],
            outputs=[],
            transform_fn=lambda _info, _inputs: None,
        )
    ]


def test_no_dep_specified():
    with pytest.raises(DagsterInvalidDefinitionError, message='Dependency must be specified'):
        PipelineDefinition(solids=solid_a_b_list(), dependencies={})


def test_circular_dep():
    with pytest.raises(DagsterInvalidDefinitionError, message='Circular reference'):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'B': {
                'b_input': DependencyDefinition('B')
            }},
        )


def test_from_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        message='Solid NOTTHERE in dependency dictionary not found',
    ):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'NOTTHERE': {
                'b_input': DependencyDefinition('A')
            }},
        )


def test_from_non_existant_input():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        message='Solid B does not have input not_an_input',
    ):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'B': {
                'not_an_input': DependencyDefinition('A')
            }},
        )


def test_to_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        message='Solid NOTTHERE in DependencyDefinition not found in solid list',
    ):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'B': {
                'b_input': DependencyDefinition('NOTTHERE')
            }},
        )


def test_to_solid_output_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        message='Solid A does not have output NOTTHERE',
    ):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'B': {
                'b_input': DependencyDefinition('A', output='NOTTHERE')
            }},
        )



import jinja2

from dagster import (
    ConfigDefinition,
    Field,
    InputDefinition,
    OutputDefinition,
    Result,
    SolidDefinition,
    check,
    types,
)

from .common import execute_sql_text_on_context


def create_templated_sql_transform_solid(name, sql, table_arguments, dependant_solids=None):
    check.str_param(name, 'name')
    check.str_param(sql, 'sql')
    check.list_param(table_arguments, 'table_arguments', of_type=str)

    dependant_solids = check.opt_list_param(
        dependant_solids, 'dependant_solids', of_type=SolidDefinition
    )

    field_dict = {}
    for table in table_arguments:
        field_dict[table] = Field(types.String)

    return SolidDefinition(
        name=name,
        inputs=[InputDefinition(solid.name) for solid in dependant_solids],
        config_def=ConfigDefinition.config_dict(field_dict),
        transform_fn=_create_templated_sql_transform_with_output(sql),
        outputs=[OutputDefinition()],
    )


def _render_template_string(template_text, config_dict):
    template = jinja2.Environment(loader=jinja2.BaseLoader).from_string(template_text)
    return template.render(**config_dict)


def _create_templated_sql_transform_with_output(sql):
    def do_transform(info, _inputs):
        rendered_sql = _render_template_string(sql, info.config)
        execute_sql_text_on_context(info.context, rendered_sql)
        yield Result(info.config)

    return do_transform



from collections import namedtuple
from enum import Enum
import imp
import importlib
import os

import click

from dagster import (
    PipelineDefinition,
    RepositoryDefinition,
    check,
)

from dagster.utils import load_yaml_from_path

INFO_FIELDS = set(
    [
        'repository_yaml',
        'pipeline_name',
        'python_file',
        'pipeline_fn_name',
        'module_name',
        'repository_fn_name',
    ]
)

PipelineTargetInfo = namedtuple(
    'PipelineTargetInfo',
    'repository_yaml pipeline_name python_file pipeline_fn_name module_name repository_fn_name',
)


class RepositoryTargetMode(Enum):
    YAML_FILE = 0
    MODULE = 1
    FILE = 2


RepositoryTargetInfo = namedtuple(
    'RepositoryTargetInfo',
    'repository_yaml module_name python_file fn_name',
)


class PipelineTargetMode(Enum):
    PIPELINE_PYTHON_FILE = 0
    PIPELINE_MODULE = 1
    REPOSITORY_PYTHON_FILE = 2
    REPOSITORY_MODULE = 3
    REPOSITORY_YAML_FILE = 4


FileTargetFunction = namedtuple('FileTargetFunction', 'python_file fn_name')
ModuleTargetFunction = namedtuple('ModuleTargetFunction', 'module_name fn_name')

RepositoryPythonFileData = namedtuple(
    'RepositoryPythonFileData',
    'file_target_function pipeline_name',
)

RepositoryModuleData = namedtuple(
    'RepositoryModuleData',
    'module_target_function pipeline_name',
)

RepositoryYamlData = namedtuple(
    'RepositoryYamlData',
    'repository_yaml pipeline_name',
)

PipelineLoadingModeData = namedtuple('PipelineLoadingModeData', 'mode data')

RepositoryLoadingModeData = namedtuple('RepositoryLoadingModeData', 'mode data')


class InvalidPipelineLoadingComboError(Exception):
    pass


class InvalidRepositoryLoadingComboError(Exception):
    pass


def check_info_fields(info, *fields):
    check.inst_param(info, 'info', PipelineTargetInfo)
    check.tuple_param(fields, 'fields')

    info_dict = info._asdict()
    for field in fields:
        if info_dict[field] is None:
            return False

    for none_field in INFO_FIELDS.difference(set(fields)):
        if info_dict[none_field] is not None:
            raise InvalidPipelineLoadingComboError()

    return True


def repo_load_invariant(condition):
    if not condition:
        raise InvalidRepositoryLoadingComboError()


def create_repository_loading_mode_data(info):
    check.inst_param(info, 'info', RepositoryTargetInfo)

    if info.repository_yaml:
        repo_load_invariant(info.module_name is None)
        repo_load_invariant(info.python_file is None)
        repo_load_invariant(info.fn_name is None)
        return RepositoryLoadingModeData(
            mode=RepositoryTargetMode.YAML_FILE,
            data=info.repository_yaml,
        )
    elif info.module_name and info.fn_name:
        repo_load_invariant(info.repository_yaml is None)
        repo_load_invariant(info.python_file is None)
        return RepositoryLoadingModeData(
            mode=RepositoryTargetMode.MODULE,
            data=ModuleTargetFunction(
                module_name=info.module_name,
                fn_name=info.fn_name,
            )
        )
    elif info.python_file and info.fn_name:
        repo_load_invariant(info.repository_yaml is None)
        repo_load_invariant(info.module_name is None)
        return RepositoryLoadingModeData(
            mode=RepositoryTargetMode.FILE,
            data=FileTargetFunction(
                python_file=info.python_file,
                fn_name=info.fn_name,
            )
        )
    else:
        raise InvalidRepositoryLoadingComboError()


def create_pipeline_loading_mode_data(info):
    check.inst_param(info, 'info', PipelineTargetInfo)

    if check_info_fields(info, 'python_file', 'repository_fn_name', 'pipeline_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY_PYTHON_FILE,
            data=RepositoryPythonFileData(
                file_target_function=FileTargetFunction(
                    python_file=info.python_file,
                    fn_name=info.repository_fn_name,
                ),
                pipeline_name=info.pipeline_name,
            )
        )
    elif check_info_fields(info, 'module_name', 'repository_fn_name', 'pipeline_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY_MODULE,
            data=RepositoryModuleData(
                module_target_function=ModuleTargetFunction(
                    module_name=info.module_name,
                    fn_name=info.repository_fn_name,
                ),
                pipeline_name=info.pipeline_name,
            )
        )
    elif check_info_fields(info, 'python_file', 'pipeline_fn_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.PIPELINE_PYTHON_FILE,
            data=FileTargetFunction(
                python_file=info.python_file,
                fn_name=info.pipeline_fn_name,
            )
        )
    elif check_info_fields(info, 'module_name', 'pipeline_fn_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.PIPELINE_MODULE,
            data=ModuleTargetFunction(
                module_name=info.module_name,
                fn_name=info.pipeline_fn_name,
            )
        )
    elif info.pipeline_name:
        for none_field in ['python_file', 'pipeline_fn_name', 'module_name', 'repository_fn_name']:
            if getattr(info, none_field) is not None:
                raise InvalidPipelineLoadingComboError(
                    '{none_field} is not None. Got {value}'.format(
                        none_field=none_field,
                        value=getattr(info, none_field),
                    )
                )

        check.invariant(info.repository_yaml is not None)

        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY_YAML_FILE,
            data=RepositoryYamlData(
                repository_yaml=info.repository_yaml,
                pipeline_name=info.pipeline_name,
            )
        )
    else:
        raise InvalidPipelineLoadingComboError()


DynamicObject = namedtuple('DynamicObject', 'module fn module_name fn_name object')


def load_file_target_function(file_target_function):
    check.inst_param(file_target_function, 'file_target_function', FileTargetFunction)
    module_name = os.path.splitext(os.path.basename(file_target_function.python_file))[0]
    module = imp.load_source(module_name, file_target_function.python_file)
    fn = getattr(module, file_target_function.fn_name)
    check.is_callable(fn)
    return DynamicObject(module, fn, module_name, file_target_function.fn_name, object=fn())


def load_module_target_function(module_target_function):
    check.inst_param(module_target_function, 'module_target_function', ModuleTargetFunction)
    module = importlib.import_module(module_target_function.module_name)
    fn = getattr(module, module_target_function.fn_name)
    check.is_callable(fn)
    return DynamicObject(
        module, fn, module_target_function.module_name, module_target_function.fn_name, object=fn()
    )


EMPHERMAL_NAME = '<<unnamed>>'


def load_repository_object_from_target_info(info):
    check.inst_param(info, 'info', RepositoryTargetInfo)

    mode_data = create_repository_loading_mode_data(info)

    if mode_data.mode == RepositoryTargetMode.YAML_FILE:
        return load_repository_from_file(mode_data.data)
    elif mode_data.mode == RepositoryTargetMode.MODULE:
        dynamic_obj = load_module_target_function(mode_data.data)
        return ensure_in_repo(dynamic_obj)
    elif mode_data.mode == RepositoryTargetMode.FILE:
        dynamic_obj = load_file_target_function(mode_data.data)
        return ensure_in_repo(dynamic_obj)
    else:
        check.failed('should not reach')


def ensure_in_repo(dynamic_obj):
    if isinstance(dynamic_obj.object, RepositoryDefinition):
        return dynamic_obj

    pipeline = dynamic_obj.object
    repo_fn = lambda: RepositoryDefinition(
        name=EMPHERMAL_NAME, pipeline_dict={pipeline.name: lambda: pipeline}
    )
    return DynamicObject(
        module=dynamic_obj.module,
        fn=repo_fn,
        module_name=dynamic_obj.module_name,
        fn_name=dynamic_obj.fn_name,
        object=repo_fn(),
    )


def load_repository_from_target_info(info):
    return check.inst(load_repository_object_from_target_info(info).fn(), RepositoryDefinition)


def load_pipeline_from_target_info(info):
    check.inst_param(info, 'info', PipelineTargetInfo)

    mode_data = create_pipeline_loading_mode_data(info)

    if mode_data.mode == PipelineTargetMode.REPOSITORY_PYTHON_FILE:
        repository = check.inst(
            load_file_target_function(mode_data.data.file_target_function).object,
            RepositoryDefinition,
        )
        return repository.get_pipeline(mode_data.data.pipeline_name)
    elif mode_data.mode == PipelineTargetMode.REPOSITORY_MODULE:
        repository = check.inst(
            load_module_target_function(mode_data.data.module_target_function).object,
            RepositoryDefinition,
        )
        return repository.get_pipeline(mode_data.data.pipeline_name)
    elif mode_data.mode == PipelineTargetMode.PIPELINE_PYTHON_FILE:
        return check.inst(
            load_file_target_function(mode_data.data).object,
            PipelineDefinition,
        )
    elif mode_data.mode == PipelineTargetMode.PIPELINE_MODULE:
        return check.inst(
            load_module_target_function(mode_data.data).object,
            PipelineDefinition,
        )
    elif mode_data.mode == PipelineTargetMode.REPOSITORY_YAML_FILE:
        repository = check.inst(
            load_repository_from_file(mode_data.data.repository_yaml).object,
            RepositoryDefinition,
        )
        return repository.get_pipeline(mode_data.data.pipeline_name)
    else:
        check.failed('Should never reach')


def load_repository_from_file(file_path):
    check.str_param(file_path, 'file_path')

    config = load_yaml_from_path(file_path)
    repository_config = check.dict_elem(config, 'repository')
    module_name = check.opt_str_elem(repository_config, 'module')
    file_name = check.opt_str_elem(repository_config, 'file')
    fn_name = check.str_elem(repository_config, 'fn')

    if module_name:
        return load_module_target_function(ModuleTargetFunction(module_name, fn_name))
    else:
        # rebase file in config off of the path in the config file
        file_name = os.path.join(os.path.dirname(os.path.abspath(file_path)), file_name)
        return load_file_target_function(FileTargetFunction(file_name, fn_name))


def apply_click_params(command, *click_params):
    for click_param in click_params:
        command = click_param(command)
    return command


def repository_config_argument(f):
    return apply_click_params(
        f,
        click.option(
            '--repository-yaml',
            '-y',
            type=click.STRING,
            help="Path to config file. Defaults to ./repository.yml."
        ),
        click.option(
            '--python-file',
            '-f',
            help='Specify python file where repository or pipeline function lives.'
        ),
        click.option(
            '--module-name',
            '-m',
            help='Specify module where repository or pipeline function lives'
        ),
        click.option('--fn-name', '-r', help='Function that returns either repository or pipeline'),
    )


def pipeline_target_command(f):
    # f = repository_config_argument(f)
    # nargs=-1 is used right now to make this argument optional
    # it can only handle 0 or 1 pipeline names
    # see create_pipeline_from_cli_args
    return apply_click_params(
        f,
        click.option(
            '--repository-yaml',
            '-y',
            type=click.STRING,
            help="Path to config file. Defaults to ./repository.yml."
        ),
        click.argument('pipeline_name', nargs=-1),
        click.option('--python-file', '-f'),
        click.option('--module-name', '-m'),
        click.option('--repository-fn-name', '-r'),
        click.option('--pipeline-fn-name', '-p'),
    )


def all_none(kwargs):
    for value in kwargs.values():
        if value is not None:
            return False
    return True


def load_target_info_from_kwargs(kwargs):
    check.dict_param(kwargs, 'kwargs')

    if all_none(kwargs):
        kwargs['repository_yaml'] = 'repository.yml'

    return RepositoryTargetInfo(**kwargs)


def reload_dynamic_object(dynamic_obj):
    check.inst_param(dynamic_obj, 'dynamic_obj', DynamicObject)

    module_name, fn_name = dynamic_obj.module_name, dynamic_obj.fn_name

    module = importlib.reload(dynamic_obj.module)
    fn = getattr(module, fn_name)
    check.is_callable(fn)

    return DynamicObject(
        object=fn(),
        module=module,
        fn=fn,
        module_name=module_name,
        fn_name=fn_name,
    )


def reload_pipeline_or_repo(dynamic_obj):
    return ensure_in_repo(reload_dynamic_object(dynamic_obj))


import json

from dagster.utils.test import get_temp_file_name
from dagster.utils.logging import (define_json_file_logger, DEBUG, INFO)


def test_basic_logging():
    with get_temp_file_name() as tf_name:
        logger = define_json_file_logger('foo', tf_name, DEBUG)
        logger.debug('bar')

        data = list(parse_json_lines(tf_name))

        assert len(data) == 1
        assert data[0]['name'] == 'foo'
        assert data[0]['msg'] == 'bar'


def parse_json_lines(tf_name):
    with open(tf_name) as f:
        for line in f:
            yield json.loads(line)


def test_no_double_write_diff_names():
    with get_temp_file_name() as tf_name:
        foo_logger = define_json_file_logger('foo', tf_name, DEBUG)
        baaz_logger = define_json_file_logger('baaz', tf_name, DEBUG)

        foo_logger.debug('foo message')
        baaz_logger.debug('baaz message')

        data = list(parse_json_lines(tf_name))

        assert len(data) == 2
        assert data[0]['name'] == 'foo'
        assert data[0]['msg'] == 'foo message'
        assert data[1]['name'] == 'baaz'
        assert data[1]['msg'] == 'baaz message'


# This demonstrates different, less global, behavior than the typical
# python logger. Same name does *not* mean the same logger instance
# here. Hence logger one *retains* debug level while logger_two is
# a separate instance with info level
def test_no_double_write_same_names():
    with get_temp_file_name() as tf_name:
        foo_logger_one = define_json_file_logger('foo', tf_name, DEBUG)
        foo_logger_two = define_json_file_logger('foo', tf_name, INFO)

        foo_logger_one.debug('logger one message')
        foo_logger_two.debug('logger two message')

        data = list(parse_json_lines(tf_name))
        assert len(data) == 1
        assert data[0]['name'] == 'foo'
        assert data[0]['msg'] == 'logger one message'



import os

import pytest

from dagster import config
from dagster.core.execution import execute_pipeline
from dagster.utils import script_relative_path
from dagster.cli.pipeline import (
    do_execute_command,
    print_pipeline,
)

from dagster.dagster_examples.pandas_hello_world.pipeline import (
    define_success_pipeline,
    define_failure_pipeline,
)


def test_pipeline_include():
    assert define_success_pipeline()


def test_execute_pipeline():
    pipeline = define_success_pipeline()
    environment = config.Environment(
        solids={'load_num_csv': config.Solid({
            'path': script_relative_path('num.csv')
        })},
    )

    result = execute_pipeline(pipeline, environment=environment)

    assert result.success

    assert result.result_for_solid('sum_solid').transformed_value().to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
    }

    assert result.result_for_solid('sum_sq_solid').transformed_value().to_dict('list') == {
        'num1': [1, 3],
        'num2': [2, 4],
        'sum': [3, 7],
        'sum_sq': [9, 49],
    }


def test_cli_execute():

    # currently paths in env files have to be relative to where the
    # script has launched so we have to simulate that
    cwd = os.getcwd()
    try:

        os.chdir(script_relative_path('../..'))

        do_execute_command(
            define_success_pipeline(),
            script_relative_path('../../pandas_hello_world/env.yml'),
            lambda *_args, **_kwargs: None,
        )
    finally:
        # restore cwd
        os.chdir(cwd)


def test_cli_execute_failure():

    # currently paths in env files have to be relative to where the
    # script has launched so we have to simulate that
    with pytest.raises(Exception, match='I am a programmer and I make error'):
        cwd = os.getcwd()
        try:

            os.chdir(script_relative_path('../..'))

            do_execute_command(
                define_failure_pipeline(),
                script_relative_path('../../pandas_hello_world/env.yml'),
                lambda *_args, **_kwargs: None,
            )
        finally:
            # restore cwd
            os.chdir(cwd)


def test_cli_print():
    print_pipeline(define_success_pipeline(), full=False, print_fn=lambda *_args, **_kwargs: None)
    print_pipeline(define_success_pipeline(), full=True, print_fn=lambda *_args, **_kwargs: None)
